# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

"""
Given an image, locates and parses the pixel data in QR codes.
"""

from __future__ import division
from yubioath.yubicommon.compat import byte2int
from collections import namedtuple


__all__ = ['parse_qr_codes']


Box = namedtuple('Box', ['x', 'y', 'w', 'h'])


def is_dark(color):  # If any R, G, or B value is < 200 we consider it dark.
    return any(byte2int(c) < 200 for c in color)


def buffer_matches(matched):
    return len(matched) == 5 \
        and max(matched[:2] + matched[3:]) <= matched[2] // 2 \
        and min(matched[:2] + matched[3:]) >= matched[2] // 6


def check_line(pixels):
    matching_dark = False
    matched = [0, 0, 0, 0, 0]
    for (i, pixel) in enumerate(pixels):
        if is_dark(pixel):  # Dark pixel
            if matching_dark:
                matched[-1] += 1
            else:
                matched = matched[1:] + [1]
                matching_dark = True
        else:  # Light pixel
            if not matching_dark:
                matched[-1] += 1
            else:
                if buffer_matches(matched):
                    width = sum(matched)
                    yield i - width, width
                matched = matched[1:] + [1]
                matching_dark = False

    # Check final state of buffer
    if matching_dark and buffer_matches(matched):
        width = sum(matched)
        yield i - width, width


def check_row(line, bpp, x_offs, x_width):
    return check_line([line[i*bpp:(i+1)*bpp]
                       for i in range(x_offs, x_offs + x_width)])


def check_col(image, bpp, x, y_offs, y_height):
    return check_line([image.scanLine(i)[x*bpp:(x+1)*bpp]
                       for i in range(y_offs, y_offs + y_height)])


def read_line(line, bpp, x_offs, x_width):
    matching_dark = not is_dark(line[x_offs*bpp:(x_offs+1)*bpp])
    matched = []
    for x in range(x_offs, x_offs + x_width):
        pixel = line[x*bpp:(x+1)*bpp]
        if is_dark(pixel):  # Dark pixel
            if matching_dark:
                matched[-1] += 1
            else:
                matched.append(1)
                matching_dark = True
        else:  # Light pixel
            if not matching_dark:
                matched[-1] += 1
            else:
                matched.append(1)
                matching_dark = False
    return matching_dark, matched


def read_bits(image, bpp, img_x, img_y, img_w, img_h, size):
    qr_x_w = img_w / size
    qr_y_h = img_h / size
    qr_data = []
    for qr_y in range(size):
        y = img_y + int(qr_y_h / 2 + qr_y * qr_y_h)
        img_line = image.scanLine(y)
        qr_line = []
        for qr_x in range(size):
            x = img_x + int(qr_x_w / 2 + qr_x * qr_x_w)
            qr_line.append(is_dark(img_line[x * bpp:(x+1) * bpp]))
        qr_data.append(qr_line)
    return qr_data


FINDER = [
    [True, True, True, True, True, True, True],
    [True, False, False, False, False, False, True],
    [True, False, True, True, True, False, True],
    [True, False, True, True, True, False, True],
    [True, False, True, True, True, False, True],
    [True, False, False, False, False, False, True],
    [True, True, True, True, True, True, True]
]


def parse_qr_codes(image, min_res=2):
    size = image.size()
    bpp = image.bytesPerLine() // size.width()

    finders = locate_finders(image, min_res)

    # Arrange finders into QR codes and extract data
    for (tl, tr, bl) in identify_groups(finders):
        min_x = min(tl.x, bl.x)
        min_y = min(tl.y, tr.y)
        width = tr.x + tr.w - min_x
        height = bl.y + bl.h - min_y

        # Determine resolution by reading timing pattern
        line = image.scanLine(min_y + int(6.5 / 7 * max(tl.h, tr.h)))
        _, line_data = read_line(line, bpp, min_x, width)
        size = len(line_data) + 12

        # Read QR code data
        yield read_bits(image, bpp, min_x, min_y, width, height, size)


def locate_finders(image, min_res):
    size = image.size()
    bpp = image.bytesPerLine() // size.width()
    finders = set()
    for y in range(0, size.height(), min_res * 3):
        for (x, w) in check_row(image.scanLine(y), bpp, 0, size.width()):
            x_offs = x + w // 2
            y_offs = max(0, y - w)
            y_height = min(size.height() - y_offs, 2 * w)
            match = next(check_col(image, bpp, x_offs, y_offs, y_height), None)
            if match:
                (pos, h) = match
                y2 = y_offs + pos
                if read_bits(image, bpp, x, y2, w, h, 7) == FINDER:
                    finders.add(Box(x, y2, w, h))
    return list(finders)


def identify_groups(locators):
    # Find top left
    for tl in locators:
        x_tol = tl.w / 14
        y_tol = tl.h / 14

        # Find top right
        for tr in locators:
            if tr.x > tl.x and abs(tl.y - tr.y) <= y_tol:

                # Find bottom left
                for bl in locators:
                    if bl.y > tl.y and abs(tl.x - bl.x) <= x_tol:
                        yield tl, tr, bl



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from __future__ import print_function

from yubioath.yubicommon.compat import int2byte, byte2int
from Crypto.Hash import HMAC, SHA
from Crypto.Protocol.KDF import PBKDF2
from Crypto.Random import get_random_bytes
try:
    from urlparse import urlparse, parse_qs
    from urllib import unquote
except ImportError:
    from urllib.parse import unquote, urlparse, parse_qs
import subprocess
import struct
import time
import sys
import re

__all__ = [
    'hmac_sha1',
    'derive_key',
    'der_pack',
    'der_read',
    'get_random_bytes'
]


#
# OATH related
#

def hmac_sha1(secret, message):
    return HMAC.new(secret, message, digestmod=SHA).digest()


def time_challenge(t=None):
    return struct.pack('>q', int((t or time.time())/30))


def parse_full(resp):
    offs = byte2int(resp[-1]) & 0xf
    return parse_truncated(resp[offs:offs+4])


def parse_truncated(resp):
    return struct.unpack('>I', resp)[0] & 0x7fffffff


def format_code(code, digits=6):
    return ('%%0%dd' % digits) % (code % 10 ** digits)


def parse_uri(uri):
    uri = uri.strip()  # Remove surrounding whitespace
    parsed = urlparse(uri)
    if parsed.scheme != 'otpauth':  # Not a uri, assume secret.
        return {'secret': uri}
    params = dict((k, v[0]) for k, v in parse_qs(parsed.query).items())
    params['name'] = unquote(parsed.path)[1:]  # Unquote and strip leading /
    params['type'] = parsed.hostname
    if 'issuer' in params and not params['name'].startswith(params['issuer']):
        params['name'] = params['issuer'] + ':' + params['name']
    return params


#
# General utils
#

def derive_key(salt, passphrase):
    if not passphrase:
        return None
    return PBKDF2(passphrase, salt, 16, 1000)


def der_pack(*values):
    return b''.join([int2byte(t) + int2byte(len(v)) + v for t, v in zip(
        values[0::2], values[1::2])])


def der_read(der_data, expected_t=None):
    t = byte2int(der_data[0])
    if expected_t is not None and expected_t != t:
        raise ValueError('Wrong tag. Expected: %x, got: %x' % (expected_t, t))
    l = byte2int(der_data[1])
    offs = 2
    if l > 0x80:
        n_bytes = l - 0x80
        l = b2len(der_data[offs:offs + n_bytes])
        offs = offs + n_bytes
    v = der_data[offs:offs + l]
    rest = der_data[offs + l:]
    if expected_t is None:
        return t, v, rest
    return v, rest


def b2len(bs):
    l = 0
    for b in bs:
        l *= 256
        l += byte2int(b)
    return l


def kill_scdaemon():
    try:
        # Works for Windows.
        from win32com.client import GetObject
        from win32api import OpenProcess, CloseHandle, TerminateProcess
        WMI = GetObject('winmgmts:')
        ps = WMI.InstancesOf('Win32_Process')
        for p in ps:
            if p.Properties_('Name').Value == 'scdaemon.exe':
                pid = p.Properties_('ProcessID').Value
                print("Killing", pid)
                handle = OpenProcess(1, False, pid)
                TerminateProcess(handle, -1)
                CloseHandle(handle)
    except ImportError:
        # Works for Linux and OS X.
        pids = subprocess.check_output(
            "ps ax | grep scdaemon | grep -v grep | awk '{ print $1 }'",
            shell=True).strip()
        if pids:
            for pid in pids.split():
                print("Killing", pid)
                subprocess.call(['kill', '-9', pid])


NON_CCID_YK_PIDS = set([0x0110, 0x0113, 0x0114, 0x0401, 0x0402, 0x0403])


def ccid_supported_but_disabled():
    """
    Check whether the first connected YubiKey supports CCID, but has it disabled.
    """

    if sys.platform in ['win32', 'cygwin']:
        pids = _get_pids_win()
    elif sys.platform == 'darwin':
        pids = _get_pids_osx()
    else:
        pids = _get_pids_linux()

    return bool(NON_CCID_YK_PIDS.intersection(pids))


def _get_pids_linux():
    pid_pattern = re.compile(r' 1050:([0-9a-f]{4}) ')
    pids = []
    for line in subprocess.check_output('lsusb').splitlines():
        match = pid_pattern.search(line.decode('ascii'))
        if match:
            pids.append(int(match.group(1), 16))
    return pids


def _get_pids_osx():
    pids = []
    vid_ok = False
    pid = None
    output = subprocess.check_output(['system_profiler', 'SPUSBDataType'])
    for line in output.splitlines():
        line = line.decode().strip()
        if line.endswith(':'):  # New entry
            if vid_ok and pid is not None:
                pids.append(pid)
            vid_ok = False
            pid = None
        if line.startswith('Vendor ID: '):
            vid_ok = line.endswith(' 0x1050')
        elif line.startswith('Product ID:'):
            pid = int(line.rsplit(' ', 1)[1], 16)

    return pids


def _get_pids_win():
    pid_pattern = re.compile(r'PID_([0-9A-F]{4})')
    pids = []
    startupinfo = subprocess.STARTUPINFO()
    startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
    output = subprocess.check_output(['wmic', 'path',
                                      'Win32_USBControllerDevice', 'get', '*'],
                                     startupinfo=startupinfo)
    for line in output.splitlines():
        if 'VID_1050' in line:
            match = pid_pattern.search(line.decode('ascii'))
            if match:
                pids.append(int(match.group(1), 16))
    return pids



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

"""
Strings for Yubico Authenticator.

Note: String names must not start with underscore (_).

"""

organization = "Yubico"
domain = "yubico.com"
app_name = "Yubico Authenticator"
win_title_1 = "Yubico Authenticator (%s)"
about_1 = "About: %s"
copyright = "Copyright &copy; Yubico"
libraries = "Library versions"
version_1 = "Version: %s"
wait = "Please wait..."
error = "Error"
menu_file = "&File"
menu_help = "&Help"
action_about = "&About"
action_add = "&Add..."
action_password = "Set/Change &password"
action_settings = "&Settings"
action_delete = "&Delete"
action_show = "&Show credentials"
action_quit = "&Quit"
password = "Password"
settings = "Settings"
advanced = "Advanced"
pass_required = "Password required"
remember = "Remember password"
no_key = "Insert a YubiKey..."
key_busy = "YubiKey already in use!"
key_present = "YubiKey found. Reading..."
key_removed = "YubiKey removed"
key_removed_desc = "There was an error communicating with the device!"
ykstd_slots = "YubiKey standard slots"
enable_slot_1 = "Read from slot %d"
n_digits = "Number of digits"
enable_systray = "Show in system tray"
kill_scdaemon = "Kill scdaemon on show"
reader_name = "Card reader name"
no_creds = "No credentials available"
add_cred = "New credential"
cred_name = "Credential name"
cred_key = "Secret key (base32)"
cred_type = "Credential type"
cred_totp = "Time based (TOTP)"
cred_hotp = "Counter based (HOTP)"
invalid_name = "Invalid name"
invalid_name_desc = "Name must be at least 3 characters"
invalid_key = "Invalid key"
invalid_key_desc = "Key must be base32 encoded"
set_pass = "Set password"
new_pass = "New password (blank for none)"
ver_pass = "Verify new password"
pass_mismatch = "Passwords do not match"
pass_mismatch_desc = "Please enter the same password twice"
touch_title = "Touch required"
touch_desc = "Touch your YubiKey now"
delete_title = "Confirm credential deletion"
delete_desc_1 = """<span>Are you sure you want to delete the credential?</span>
<br>
This action cannot be undone.
<br><br>
<b>Delete credential: %s</b>
"""
slot = "YubiKey slot"
slot_2 = "Slot %d (%s)"
free = "free"
in_use = "in use"
require_touch = "Require touch"
no_slot = "No slot chosen"
no_slot_desc = "Please choose a slot to write the credential to"
overwrite_slot = "Overwrite slot?"
overwrite_slot_desc_1 = "This will overwrite the credential currently stored " \
    "in slot %d. This action cannot be undone."
overwrite_entry = "Overwrite entry?"
overwrite_entry_desc = "An entry with this username already exists.\n\nDo you wish to overwrite it? This action cannot be undone."
qr_scan = "Scan a QR code"
qr_scanning = "Scanning for QR code..."
qr_not_found = "QR code not found"
qr_not_found_desc = "No usable QR code detected. Make sure the QR code is " \
    "fully visible on your primary screen and try again."
qr_not_supported = "Credential not supported"
qr_not_supported_desc = "This credential type is not supported for slot " \
    "based usage."
tt_slot_enabled_1 = "Check to calculate TOTP codes using the YubiKey " \
    "standard slot %d credential."
tt_num_digits = "The number of digits to show for the credential."
tt_systray = "When checked, display an icon in the systray, and leave the " \
    "application running there when closed."
tt_kill_scdaemon = "Kills any running scdaemon process when the window is " \
    "shown. This is useful when using this application together with GnuPG " \
    "to avoid GnuPG locking the device."
tt_reader_name = "Changes the default smartcard reader name to look for. " \
    "This can be used to target a specific YubiKey when multiple are used, " \
    "or to target an NFC reader."
ccid_disabled = '<b>CCID (smart card capabilities) is disabled on the ' \
    'inserted YubiKey.</b><br><br>Without CCID enabled, you will only be ' \
    'able to store 2 credentials.<br><br>' \
    '<a href="%s">Learn how to enable CCID</a><br>'
no_space="No space available"
no_space_desc="There is not enough space to add another credential on your device.\n\nTo create free space to add a new credential, delete those you no longer need."

def _translate(qt):
    values = globals()
    for key, value in values.items():
        if isinstance(value, basestring) and not key.startswith('_'):
            values[key] = qt.tr(value)



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from __future__ import absolute_import

from PySide import QtCore, QtGui
from functools import wraps
from inspect import getargspec

__all__ = ['get_text', 'get_active_window', 'is_minimized', 'connect_once']


class _DefaultMessages(object):

    def __init__(self, default_m, m=None):
        self._defaults = default_m
        self._m = m

    def __getattr__(self, method_name):
        if hasattr(self._m, method_name):
            return getattr(self._m, method_name)
        else:
            return getattr(self._defaults, method_name)

def default_messages(_m, name='m'):
    def inner(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            index = getargspec(fn).args.index(name)
            if len(args) > index:
                args = list(args)
                args[index] = _DefaultMessages(_m, args[index])
            else:
                kwargs[name] = _DefaultMessages(_m, kwargs.get(name))
            return fn(*args, **kwargs)
        return wrapper
    return inner


def get_text(*args, **kwargs):
    flags = (
        QtCore.Qt.WindowTitleHint |
        QtCore.Qt.WindowSystemMenuHint
    )
    kwargs['flags'] = flags
    return QtGui.QInputDialog.getText(*args, **kwargs)


def get_active_window():
    active_win = QtGui.QApplication.activeWindow()
    if active_win is not None:
        return active_win

    wins = [w for w in QtGui.QApplication.topLevelWidgets()
            if isinstance(w, QtGui.QDialog) and w.isVisible()]

    if not wins:
        return QtCore.QCoreApplication.instance().window

    return wins[0]  # TODO: If more than one candidates remain, find best one.


def connect_once(signal, slot):
    def wrapped(*args, **kwargs):
        signal.disconnect(wrapped)
        slot(*args, **kwargs)
    signal.connect(wrapped)


def is_minimized(window):
    """Returns True iff the window is minimized or has been sent to the tray"""
    return not window.isVisible() or window.isMinimized()



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from __future__ import absolute_import

from PySide import QtGui, QtCore
from .worker import Worker
import os
import sys
import importlib
from .. import compat
from .osx import app_services

__all__ = ['Application', 'Dialog', 'MutexLocker']

TOP_SECTION = '<b>%s</b>'
SECTION = '<br><b>%s</b>'


class Dialog(QtGui.QDialog):

    def __init__(self, *args, **kwargs):
        super(Dialog, self).__init__(*args, **kwargs)
        self.setWindowFlags(self.windowFlags() ^
                            QtCore.Qt.WindowContextHelpButtonHint)
        self._headers = _Headers()

    @property
    def headers(self):
        return self._headers

    def section(self, title):
        return self._headers.section(title)


class _Headers(object):

    def __init__(self):
        self._first = True

    def section(self, title):
        if self._first:
            self._first = False
            section = TOP_SECTION % title
        else:
            section = SECTION % title
        return QtGui.QLabel(section)


class _MainWindow(QtGui.QMainWindow):

    def __init__(self):
        super(_MainWindow, self).__init__()
        self._widget = None

    def hide(self):
        if sys.platform == 'darwin':
            app_services.osx_hide()
        else:
            super(_MainWindow, self).hide()

    def customEvent(self, event):
        event.callback()
        event.accept()


class Application(QtGui.QApplication):
    _quit = False

    def __init__(self, m=None, version=None):
        super(Application, self).__init__(sys.argv)
        self._determine_basedir()
        self._read_package_version(version)

        self.window = _MainWindow()

        if m:  # Run all strings through Qt translation
            for key in dir(m):
                if (isinstance(key, compat.string_types) and
                        not key.startswith('_')):
                    setattr(m, key, self.tr(getattr(m, key)))

        self.worker = Worker(self.window, m)

    def _determine_basedir(self):
        if getattr(sys, 'frozen', False):
            # we are running in a PyInstaller bundle
            self.basedir = sys._MEIPASS
        else:
            # we are running in a normal Python environment
            top_module_str = __package__.split('.')[0]
            top_module = importlib.import_module(top_module_str)
            self.basedir = os.path.dirname(top_module.__file__)

    def _read_package_version(self, version):
        if version is None:
            return

        pversion_fn = os.path.join(self.basedir, 'package_version.txt')
        try:
            with open(pversion_fn, 'r') as f:
                pversion = int(f.read().strip())
        except:
            pversion = 0

        if pversion > 0:
            version += '.%d' % pversion

        self.version = version

    def ensure_singleton(self, name=None):
        if not name:
            name = self.applicationName()
        from PySide import QtNetwork
        self._l_socket = QtNetwork.QLocalSocket()
        self._l_socket.connectToServer(name, QtCore.QIODevice.WriteOnly)
        if self._l_socket.waitForConnected():
            self._stop()
            sys.exit(0)
        else:
            self._l_server = QtNetwork.QLocalServer()
            if not self._l_server.listen(name):
                QtNetwork.QLocalServer.removeServer(name)
                self._l_server.listen(name)
            self._l_server.newConnection.connect(self._show_window)

    def _show_window(self):
        self.window.show()
        self.window.activateWindow()

    def quit(self):
        super(Application, self).quit()
        self._quit = True

    def _stop(self):
        worker_thread = self.worker.thread()
        worker_thread.quit()
        worker_thread.wait()
        self.deleteLater()
        sys.stdout.flush()
        sys.stderr.flush()

    def exec_(self):
        if not self._quit:
            status = super(Application, self).exec_()
        else:
            status = 0
        self._stop()
        return status


class MutexLocker(object):

    """Drop-in replacement for QMutexLocker that can start unlocked."""

    def __init__(self, mutex, lock=True):
        self._mutex = mutex
        self._locked = False
        if lock:
            self.relock()

    def lock(self, try_lock=False):
        if try_lock:
            self._locked = self._mutex.tryLock()
        else:
            self._mutex.lock()
            self._locked = True
        return self._locked and self or None

    def relock(self):
        self.lock()

    def unlock(self):
        if self._locked:
            self._mutex.unlock()

    def __del__(self):
        self.unlock()



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from getpass import getpass
from yubioath.core.exc import CardError
import sys


class CliUnlocker(object):

    def __init__(self, keystore, save=False):
        self.keystore = keystore
        self._save = save

    def __call__(self, device):
        key = self.keystore.get(device.id)
        if key:
            try:
                device.unlock(key)
            except CardError:
                sys.stderr.write('Incorrect password from file.\n')
                self.keystore.delete(device.id)

        while device.locked:
            pw = getpass('Password: ')
            key = device.calculate_key(pw)
            try:
                device.unlock(key)
                if self._save:
                    self.keystore.put(device.id, key)
                    sys.stderr.write('Password saved to %s\n' %
                                     self.keystore.fname)
            except CardError:
                sys.stderr.write('Incorrect password!\n')



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from .exc import CardError, InvalidSlotError, NeedsTouchError
from .utils import (format_code, parse_full, time_challenge)

YKLEGACY_AID = b'\xa0\x00\x00\x05\x27\x20\x01'

INS_CHALRESP = 0x01

SLOTS = [
    -1,
    0x30,
    0x38
]


class LegacyOathCcid(object):

    """
    CCID interface to a legacy OATH-enabled YubiKey.
    """

    def __init__(self, device):
        self._device = device

        self._select()

    def _send(self, ins, data='', p1=0, p2=0, expected=0x9000):
        resp, status = self._device.send_apdu(0, ins, p1, p2, data)
        if expected != status:
            raise CardError(status)
        return resp

    def _select(self):
        self._send(0xa4, YKLEGACY_AID, p1=0x04)

    def calculate(self, slot, digits=6, timestamp=None, mayblock=0):
        data = time_challenge(timestamp)
        try:
            resp = self._send(INS_CHALRESP, data, p1=SLOTS[slot])
        except CardError as e:
            if e.status == 0x6985:
                raise NeedsTouchError()
            raise
        if not resp:
            raise InvalidSlotError()
        return format_code(parse_full(resp), digits)



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from PySide import QtGui, QtCore

from yubioath import __version__ as version
from yubioath.gui.view.ccid_disabled import CcidDisabledDialog
from yubioath.yubicommon import qt
from ..cli.keystore import CONFIG_HOME

try:
    from ..core.legacy_otp import ykpers_version
except ImportError:
    ykpers_version = 'None'
from ..core.utils import kill_scdaemon
from ..core.exc import NoSpaceError
from . import messages as m
from .controller import GuiController
from .ccid import CardStatus
from .view.systray import Systray
from .view.codes import CodesWidget
from .view.settings import SettingsDialog
from .view.add_cred import AddCredDialog
from .view.add_cred_legacy import AddCredDialog as AddCredLegacyDialog
from .view.set_password import SetPasswordDialog
import sys
import os
import signal
import argparse


ABOUT_TEXT = """
<h2>%s</h2>
%s<br>
%s
<h4>%s</h4>
%%s
<br><br>
""" % (m.app_name, m.copyright, m.version_1, m.libraries)


class MainWidget(QtGui.QStackedWidget):

    def __init__(self, controller):
        super(MainWidget, self).__init__()

        self._controller = controller

        self._build_ui()
        controller.refreshed.connect(self._refresh)
        controller.ccid_disabled.connect(self.ccid_disabled)
        controller.watcher.status_changed.connect(self._set_status)

    def ccid_disabled(self):
        if not self._controller.mute_ccid_disabled_warning:
            dialog = CcidDisabledDialog()
            dialog.exec_()
            if dialog.do_not_ask_again.isChecked():
                self._controller.mute_ccid_disabled_warning = 1

    def showEvent(self, event):
        event.accept()

    def _build_ui(self):
        self.codes_widget = CodesWidget(self._controller)
        self.no_key_widget = QtGui.QLabel(m.no_key)
        self.no_key_widget.setAlignment(QtCore.Qt.AlignCenter)
        self.addWidget(self.no_key_widget)
        self.addWidget(self.codes_widget)

    def _refresh(self):
        if self._controller.credentials is None:
            self.setCurrentIndex(0)
        else:
            self.setCurrentIndex(1)

    def _set_status(self, status):
        if status == CardStatus.NoCard:
            self.no_key_widget.setText(m.no_key)
        elif status == CardStatus.InUse:
            self.no_key_widget.setText(m.key_busy)
        elif status == CardStatus.Present:
            self.no_key_widget.setText(m.key_present)


class YubiOathApplication(qt.Application):

    def __init__(self, args):
        super(YubiOathApplication, self).__init__(m, version)

        QtCore.QCoreApplication.setOrganizationName(m.organization)
        QtCore.QCoreApplication.setOrganizationDomain(m.domain)
        QtCore.QCoreApplication.setApplicationName(m.app_name)

        self.ensure_singleton()

        self._widget = None
        self.settings = qt.Settings.wrap(
            os.path.join(CONFIG_HOME, 'settings.ini'),
            QtCore.QSettings.IniFormat)
        self._settings = self.settings.get_group('settings')

        self._controller = GuiController(self, self._settings)

        self._systray = Systray(self)

        self._init_systray(args.tray or self._settings.get('systray', False))
        self._init_window(not args.tray)

    def _init_systray(self, show=False):
        self._systray.setIcon(QtGui.QIcon(':/yubioath.png'))
        self._systray.setVisible(show)

    def _init_window(self, show=True):
        self.window.setWindowTitle(m.win_title_1 % self.version)
        self.window.setWindowIcon(QtGui.QIcon(':/yubioath.png'))
        self.window.resize(self._settings.get('size', QtCore.QSize(320, 340)))

        self._build_menu_bar()

        self.window.showEvent = self._on_shown
        self.window.closeEvent = self._on_closed
        self.window.hideEvent = self._on_hide

        if show:
            self.window.show()
            self.window.raise_()

    def _build_menu_bar(self):
        file_menu = self.window.menuBar().addMenu(m.menu_file)
        self._add_action = QtGui.QAction(m.action_add, file_menu)
        self._add_action.triggered.connect(self._add_credential)
        file_menu.addAction(self._add_action)
        self._password_action = QtGui.QAction(m.action_password, file_menu)
        self._password_action.triggered.connect(self._change_password)
        self._password_action.setEnabled(False)
        file_menu.addAction(self._password_action)
        settings_action = QtGui.QAction(m.action_settings, file_menu)
        settings_action.triggered.connect(self._show_settings)
        file_menu.addAction(settings_action)
        file_menu.addSeparator()
        quit_action = QtGui.QAction(m.action_quit, file_menu)
        quit_action.triggered.connect(self._systray.quit)
        file_menu.addAction(quit_action)

        help_menu = self.window.menuBar().addMenu(m.menu_help)
        about_action = QtGui.QAction(m.action_about, help_menu)
        about_action.triggered.connect(self._about)
        help_menu.addAction(about_action)

        self._controller.refreshed.connect(self._refresh_menu)

    def _refresh_menu(self):
        enabled = bool(self._controller._reader)
        self._password_action.setEnabled(enabled)

    def _on_shown(self, event):
        if self._settings.get('kill_scdaemon', False):
            kill_scdaemon()

        if not self._widget:
            self._widget = MainWidget(self._controller)
            self.window.setCentralWidget(self._widget)
        self._controller.refresh_codes()
        event.accept()

    def _on_hide(self, event):
        self._controller.forget_passwords()
        event.accept()

    def _on_closed(self, event):
        self._settings['size'] = self.window.size()
        if self._systray.isVisible():
            # Unless move is called the position isn't saved!
            self.window.move(self.window.pos())
            self.window.hide()
            event.ignore()
        else:
            event.accept()

    def _libversions(self):
        return 'ykpers: %s' % ykpers_version

    def _about(self):
        QtGui.QMessageBox.about(self.window, m.about_1 % m.app_name,
                                ABOUT_TEXT % (self.version, self._libversions()))

    def _add_credential(self):
        c = self._controller.get_capabilities()
        if c.ccid:
            dialog = AddCredDialog(self.worker, c.version, self._controller.get_entry_names(), parent=self.window)
            if dialog.exec_():
                if not self._controller._reader:
                    QtGui.QMessageBox.critical(
                        self.window, m.key_removed, m.key_removed_desc)
                else:
                    try:
                        self._controller.add_cred(
                        dialog.name, dialog.key, oath_type=dialog.oath_type,
                        digits=dialog.n_digits,
                        require_touch=dialog.require_touch)
                    except NoSpaceError:
                        QtGui.QMessageBox.critical(self.window, m.no_space, m.no_space_desc)
        elif c.otp:
            dialog = AddCredLegacyDialog(self.worker, c.otp, parent=self.window)
            if dialog.exec_():
                self._controller.add_cred_legacy(dialog.slot, dialog.key,
                                                 dialog.touch)
                key = 'slot%d' % dialog.slot
                self._settings[key] = dialog.n_digits
                self._controller.refresh_codes()
        else:
            QtGui.QMessageBox.critical(self.window, 'No key', 'No key')

    def _change_password(self):
        dialog = SetPasswordDialog(self.window)
        if dialog.exec_():
            if not self._controller._reader:
                QtGui.QMessageBox.critical(
                    self.window, m.key_removed, m.key_removed_desc)
            else:
                self._controller.set_password(dialog.password, dialog.remember)

    def _show_settings(self):
        if SettingsDialog(self.window, self._settings).exec_():
            self._systray.setVisible(self._settings.get('systray', False))
            self._controller.settings_changed()


def parse_args():
        parser = argparse.ArgumentParser(description='Yubico Authenticator',
                                         add_help=True)
        parser.add_argument('-t', '--tray', action='store_true', help='starts '
                            'the application minimized to the systray')
        return parser.parse_args()


def main():
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    app = YubiOathApplication(parse_args())
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from ..core.controller import Controller
from ..core.standard import YubiOathCcid
from ..core.exc import CardError
from getpass import getpass
import sys


class CliController(Controller):

    def __init__(self, keystore, save=False):
        self.keystore = keystore
        self._save = save

    def _prompt_touch(self):
        sys.stderr.write('Touch your YubiKey...\n')

    def unlock(self, device):
        key = self.keystore.get(device.id)
        if key:
            try:
                device.unlock(key)
            except CardError:
                sys.stderr.write('Incorrect password from file.\n')
                self.keystore.delete(device.id)

        while device.locked:
            pw = getpass('Password: ')
            key = device.calculate_key(pw)
            try:
                device.unlock(key)
                if self._save:
                    self.keystore.put(device.id, key)
                    sys.stderr.write('Password saved to %s\n' %
                                     self.keystore.fname)
            except CardError:
                sys.stderr.write('Incorrect password!\n')

    def set_password(self, ccid_dev, password, remember=False):
        dev = YubiOathCcid(ccid_dev)
        key = super(CliController, self).set_password(dev, password)
        if remember:
            self.keystore.put(dev.id, key)
            sys.stderr.write('Password saved to %s\n' % self.keystore.fname)
        else:
            self.keystore.delete(dev.id)

    def add_cred(self, ccid_dev, *args, **kwargs):
        dev = YubiOathCcid(ccid_dev)
        super(CliController, self).add_cred(dev, *args, **kwargs)

    def delete_cred(self, ccid_dev, name):
        dev = YubiOathCcid(ccid_dev)
        super(CliController, self).delete_cred(dev, name)

    def reset_device(self, ccid_dev):
        dev = YubiOathCcid(ccid_dev)
        self.keystore.delete(dev.id)
        super(CliController, self).reset_device(dev)



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

import os
import json
from binascii import b2a_hex, a2b_hex

CONFIG_HOME = os.path.join(os.path.expanduser('~'), '.yubioath')
KEY_FILE = os.path.join(CONFIG_HOME, 'keys.json')


def get_keystore():
    return Keystore(KEY_FILE)


def _to_hex(val):
    return b2a_hex(val).decode('ascii')


class Keystore(object):

    def __init__(self, fname):
        self.fname = fname
        self._data = {}
        if os.path.isfile(fname):
            with open(fname) as f:
                try:
                    data = json.load(f)
                    if isinstance(data, dict):
                        self._data = data
                except ValueError:
                    pass

    def get(self, id):
        key = _to_hex(id)
        if key in self._data:
            return a2b_hex(self._data[key])
        return None

    def put(self, id, key):
        if not key:
            self.delete(id)
        else:
            self._data[_to_hex(id)] = _to_hex(key)
            self._save()

    def delete(self, id):
        key = _to_hex(id)
        if key in self._data:
            del self._data[key]
            self._save()

    def _save(self):
        directory = os.path.dirname(self.fname)
        if not os.path.isdir(directory):
            os.makedirs(directory)
        with open(self.fname, 'w') as f:
            json.dump(self._data, f)

    def clear(self):
        if os.path.isfile(self.fname):
            os.remove(self.fname)



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from .ccid import CardStatus
from smartcard.Exceptions import SmartcardException
from smartcard.scard import *
from PySide import QtCore
import smartcard.util
import threading
import time


class LLScardDevice(object):

    """
    Low level pyscard based backend (Windows chokes on the high level one
    whenever you remove the key and re-insert it).
    """

    def __init__(self, context, card, protocol):
        self._context = context
        self._card = card
        self._protocol = protocol

    def send_apdu(self, cl, ins, p1, p2, data):
        apdu = [cl, ins, p1, p2, len(data)] + map(ord, data)
        hresult, response = SCardTransmit(self._card, self._protocol, apdu)
        if hresult != SCARD_S_SUCCESS:
            raise Exception('Failed to transmit: ' +
                            SCardGetErrorMessage(hresult))
        status = response[-2] << 8 | response[-1]
        return ''.join(map(chr, response[:-2])), status

    def __del__(self):
        self.close()

    def close(self):
        SCardDisconnect(self._card, SCARD_UNPOWER_CARD)
        SCardReleaseContext(self._context)


class PollerThread(threading.Thread):

    def __init__(self, watcher):
        super(PollerThread, self).__init__()
        self._watcher = watcher
        self.daemon = True
        self.running = True

    def run(self):
        old_readers = []
        while self.running:
            readers = self._list()
            added = [r for r in readers if r not in old_readers]
            removed = [r for r in old_readers if r not in readers]
            self._watcher._update(added, removed)
            old_readers = readers
            time.sleep(2)

    def _list(self):
        try:
            hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
            if hresult != SCARD_S_SUCCESS:
                raise Exception('Failed to establish context : ' +
                                SCardGetErrorMessage(hresult))

            try:
                hresult, readers = SCardListReaders(hcontext, [])
                if hresult != SCARD_S_SUCCESS:
                    raise Exception('Failed to list readers: ' +
                                    SCardGetErrorMessage(hresult))
                return readers
            finally:
                hresult = SCardReleaseContext(hcontext)
                if hresult != SCARD_S_SUCCESS:
                    raise Exception('Failed to release context: ' +
                                    SCardGetErrorMessage(hresult))
        except:
            return []


class CardWatcher(QtCore.QObject):
    status_changed = QtCore.Signal(int)

    def __init__(self, reader_name, callback, parent=None):
        super(CardWatcher, self).__init__(parent)

        self._status = CardStatus.NoCard
        self.reader_name = reader_name
        self._callback = callback or (lambda _: _)
        self._reader = None
        self._thread = PollerThread(self)
        self._thread.start()

    def _update(self, added, removed):
        if self._reader in removed:  # Device removed
            self.reader = None
            self._set_status(CardStatus.NoCard)

        if self._reader is None:
            for reader in added:
                if self.reader_name in reader:
                    self.reader = reader
                    self._set_status(CardStatus.Present)
                    return

    @property
    def status(self):
        return self._status

    def _set_status(self, value):
        if self._status != value:
            self._status = value
            self.status_changed.emit(value)

    @property
    def reader(self):
        return self._reader

    @reader.setter
    def reader(self, value):
        self._reader = value
        self._callback(self, value)

    def open(self):
        if self._reader:
            try:
                hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
                if hresult != SCARD_S_SUCCESS:
                    raise Exception('Failed to establish context : ' +
                                    SCardGetErrorMessage(hresult))

                hresult, hcard, dwActiveProtocol = SCardConnect(
                    hcontext, self._reader,
                    SCARD_SHARE_SHARED, SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1)
                if hresult != SCARD_S_SUCCESS:
                    raise Exception('Unable to connect: ' +
                                    SCardGetErrorMessage(hresult))
                return LLScardDevice(hcontext, hcard, dwActiveProtocol)
            except Exception:
                self._set_status(CardStatus.InUse)

    def __del__(self):
        self._thread.running = False
        self._thread.join()


def observe_reader(reader_name='Yubikey', callback=None):
    return CardWatcher(reader_name, callback)



# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from .libloader import load_library

__all__ = ['load_library', 'use_library', 'CLibrary']


def use_library(libname, version=None):
    lib = load_library(libname, version)

    def define(func_name, argtypes, restype=None):
        try:
            f = getattr(lib, func_name)
            f.argtypes = argtypes
            f.restype = restype
        except AttributeError:
            print("Undefined symbol: %s" % func_name)

            def error(*args, **kwargs):
                raise Exception("Undefined symbol: %s" % func_name)
            f = error
        return f
    return define


class CLibrary(object):
    """
    Base class for extending to create python wrappers for c libraries.

    Example:
        class Foo(CLibrary):
            foo_func = [c_bool, c_char_p], int

        foo = Foo('libfoo')

        assert foo.foo_func(True, 'Hello!') == 7
    """
    def __init__(self, libname, version=None):
        self._lib = use_library(libname, version)

    def __getattribute__(self, name):
        val = object.__getattribute__(self, name)
        if isinstance(val, tuple) and len(val) == 2:
            return self._lib(name, *val)
        return val



# ----------------------------------------------------------------------------
# Copyright (c) 2008 David James
# Copyright (c) 2006-2008 Alex Holkner
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

from __future__ import absolute_import

import os.path
import re
import sys
import glob
import platform
import ctypes
import ctypes.util


def _environ_path(name):
    if name in os.environ:
        return os.environ[name].split(":")
    else:
        return []


class LibraryLoader(object):

    def __init__(self):
        self.other_dirs = []

    def load_library(self, libname, version=None):
        """Given the name of a library, load it."""

        paths = self.getpaths(libname)

        for path in paths:
            if os.path.exists(path):
                return self.load(path)

        raise ImportError("%s not found." % libname)

    def load(self, path):
        """Given a path to a library, load it."""
        try:
            # Darwin requires dlopen to be called with mode RTLD_GLOBAL instead
            # of the default RTLD_LOCAL.  Without this, you end up with
            # libraries not being loadable, resulting in "Symbol not found"
            # errors
            if sys.platform == 'darwin':
                return ctypes.CDLL(path, ctypes.RTLD_GLOBAL)
            else:
                return ctypes.cdll.LoadLibrary(path)
        except OSError as e:
            raise ImportError(e)

    def getpaths(self, libname):
        """Return a list of paths where the library might be found."""
        if os.path.isabs(libname):
            yield libname
        else:
            # FIXME / TODO return '.' and os.path.dirname(__file__)
            for path in self.getplatformpaths(libname):
                yield path

            path = ctypes.util.find_library(libname)
            if path:
                yield path

    def getplatformpaths(self, libname):
        return []

# Darwin (Mac OS X)


class DarwinLibraryLoader(LibraryLoader):
    name_formats = ["lib%s.dylib", "lib%s.so", "lib%s.bundle", "%s.dylib",
                    "%s.so", "%s.bundle", "%s"]

    def getplatformpaths(self, libname):
        if os.path.pathsep in libname:
            names = [libname]
        else:
            names = [format % libname for format in self.name_formats]

        for dir in self.getdirs(libname):
            for name in names:
                yield os.path.join(dir, name)

    def getdirs(self, libname):
        '''Implements the dylib search as specified in Apple documentation:

        http://developer.apple.com/documentation/DeveloperTools/Conceptual/
            DynamicLibraries/Articles/DynamicLibraryUsageGuidelines.html

        Before commencing the standard search, the method first checks
        the bundle's ``Frameworks`` directory if the application is running
        within a bundle (OS X .app).
        '''

        dyld_fallback_library_path = _environ_path(
            "DYLD_FALLBACK_LIBRARY_PATH")
        if not dyld_fallback_library_path:
            dyld_fallback_library_path = [os.path.expanduser('~/lib'),
                                          '/usr/local/lib', '/usr/lib']

        dirs = []

        if '/' in libname:
            dirs.extend(_environ_path("DYLD_LIBRARY_PATH"))
        else:
            dirs.extend(_environ_path("LD_LIBRARY_PATH"))
            dirs.extend(_environ_path("DYLD_LIBRARY_PATH"))

        dirs.extend(self.other_dirs)
        dirs.append(".")
        dirs.append(os.path.dirname(__file__))

        if hasattr(sys, 'frozen') and sys.frozen == 'macosx_app':
            dirs.append(os.path.join(
                os.environ['RESOURCEPATH'],
                '..',
                'Frameworks'))

        if hasattr(sys, 'frozen'):
            dirs.append(sys._MEIPASS)

        dirs.extend(dyld_fallback_library_path)

        return dirs

# Posix


class PosixLibraryLoader(LibraryLoader):
    _ld_so_cache = None

    def load_library(self, libname, version=None):
        try:
            return self.load(ctypes.util.find_library(libname))
        except ImportError:
            return super(PosixLibraryLoader, self).load_library(
                libname, version)

    def _create_ld_so_cache(self):
        # Recreate search path followed by ld.so.  This is going to be
        # slow to build, and incorrect (ld.so uses ld.so.cache, which may
        # not be up-to-date).  Used only as fallback for distros without
        # /sbin/ldconfig.
        #
        # We assume the DT_RPATH and DT_RUNPATH binary sections are omitted.

        directories = []
        for name in ("LD_LIBRARY_PATH",
                     "SHLIB_PATH",  # HPUX
                     "LIBPATH",  # OS/2, AIX
                     "LIBRARY_PATH",  # BE/OS
                     ):
            if name in os.environ:
                directories.extend(os.environ[name].split(os.pathsep))
        directories.extend(self.other_dirs)
        directories.append(".")
        directories.append(os.path.dirname(__file__))

        try:
            directories.extend([dir.strip()
                               for dir in open('/etc/ld.so.conf')])
        except IOError:
            pass

        unix_lib_dirs_list = ['/lib', '/usr/lib', '/lib64', '/usr/lib64']
        if sys.platform.startswith('linux'):
            # Try and support multiarch work in Ubuntu
            # https://wiki.ubuntu.com/MultiarchSpec
            bitage = platform.architecture()[0]
            if bitage.startswith('32'):
                # Assume Intel/AMD x86 compat
                unix_lib_dirs_list += [
                    '/lib/i386-linux-gnu', '/usr/lib/i386-linux-gnu']
            elif bitage.startswith('64'):
                # Assume Intel/AMD x86 compat
                unix_lib_dirs_list += [
                    '/lib/x86_64-linux-gnu', '/usr/lib/x86_64-linux-gnu']
            else:
                # guess...
                unix_lib_dirs_list += glob.glob('/lib/*linux-gnu')
        directories.extend(unix_lib_dirs_list)

        cache = {}
        lib_re = re.compile(r'lib(.*)\.s[ol]')
        ext_re = re.compile(r'\.s[ol]$')
        for dir in directories:
            try:
                for path in glob.glob("%s/*.s[ol]*" % dir):
                    file = os.path.basename(path)

                    # Index by filename
                    if file not in cache:
                        cache[file] = path

                    # Index by library name
                    match = lib_re.match(file)
                    if match:
                        library = match.group(1)
                        if library not in cache:
                            cache[library] = path
            except OSError:
                pass

        self._ld_so_cache = cache

    def getplatformpaths(self, libname):
        if self._ld_so_cache is None:
            self._create_ld_so_cache()

        result = self._ld_so_cache.get(libname)
        if result:
            yield result

        path = ctypes.util.find_library(libname)
        if path:
            yield os.path.join("/lib", path)

# Windows


class _WindowsLibrary(object):

    def __init__(self, path):
        self.cdll = ctypes.cdll.LoadLibrary(path)
        self.windll = ctypes.windll.LoadLibrary(path)

    def __getattr__(self, name):
        try:
            return getattr(self.cdll, name)
        except AttributeError:
            try:
                return getattr(self.windll, name)
            except AttributeError:
                raise


class WindowsLibraryLoader(LibraryLoader):
    name_formats = ["%s.dll", "lib%s.dll", "%slib.dll"]

    def load_library(self, libname, version=None):
        try:
            result = LibraryLoader.load_library(self, libname, version)
        except ImportError:
            result = None
            if os.path.sep not in libname:
                formats = self.name_formats[:]
                if version:
                    formats.append("lib%%s-%s.dll" % version)
                for name in formats:
                    try:
                        result = getattr(ctypes.cdll, name % libname)
                        if result:
                            break
                    except WindowsError:
                        result = None
            if result is None:
                try:
                    result = getattr(ctypes.cdll, libname)
                except WindowsError:
                    result = None
            if result is None:
                raise ImportError("%s not found." % libname)
        return result

    def load(self, path):
        return _WindowsLibrary(path)

    def getplatformpaths(self, libname):
        if os.path.sep not in libname:
            for name in self.name_formats:
                dll_in_current_dir = os.path.abspath(name % libname)
                if os.path.exists(dll_in_current_dir):
                    yield dll_in_current_dir
                path = ctypes.util.find_library(name % libname)
                if path:
                    yield path

# Platform switching

# If your value of sys.platform does not appear in this dict, please contact
# the Ctypesgen maintainers.

loaderclass = {
    "darwin":   DarwinLibraryLoader,
    "cygwin":   WindowsLibraryLoader,
    "win32":    WindowsLibraryLoader
}

loader = loaderclass.get(sys.platform, PosixLibraryLoader)()


def add_library_search_dirs(other_dirs):
    loader.other_dirs = other_dirs

load_library = loader.load_library

del loaderclass



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

import ctypes
from ..ctypes import CLibrary

__all__ = ['app_services']


class ProcessSerialNumber(ctypes.Structure):
    _fields_ = [('highLongOfPsn', ctypes.c_uint32),
                ('lowLongOfPSN', ctypes.c_uint32)]


class ApplicationServices(CLibrary):
    ShowHideProcess = [ctypes.POINTER(ProcessSerialNumber), ctypes.c_bool], None
    GetFrontProcess = [ctypes.POINTER(ProcessSerialNumber)], None

    def osx_hide(self):
        """ Hide the window and let the dock
        icon be able to show the window again. """
        psn = ProcessSerialNumber()
        self.GetFrontProcess(ctypes.byref(psn))
        self.ShowHideProcess(ctypes.byref(psn), False)

app_services = ApplicationServices('ApplicationServices')



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from ..core.ccid import ScardDevice
from smartcard import System
from smartcard.ReaderMonitoring import ReaderMonitor, ReaderObserver
from smartcard.CardMonitoring import CardMonitor, CardObserver
from smartcard.Exceptions import SmartcardException
from smartcard.pcsc.PCSCExceptions import EstablishContextException
from PySide import QtCore
import weakref


class _CcidReaderObserver(ReaderObserver):

    def __init__(self, controller):
        self._controller = weakref.ref(controller)
        self._monitor = ReaderMonitor()
        self._monitor.addObserver(self)

    def update(self, observable, tup):
        (added, removed) = tup
        c = self._controller()
        if c:
            c._update(added, removed)

    def delete(self):
        self._monitor.deleteObservers()


class _CcidCardObserver(CardObserver):

    def __init__(self, controller):
        self._controller = weakref.ref(controller)
        self._monitor = CardMonitor()
        self._monitor.addObserver(self)

    def update(self, observable, tup):
        (added, removed) = tup
        c = self._controller()
        if c:
            c._update([card.reader for card in added],
                      [r.reader for r in removed])

    def delete(self):
        self._monitor.deleteObservers()


class CardStatus:
    NoCard, InUse, Present = range(3)


class CardWatcher(QtCore.QObject):
    status_changed = QtCore.Signal(int)

    def __init__(self, reader_name, callback, parent=None):
        super(CardWatcher, self).__init__(parent)

        self._status = CardStatus.NoCard
        self.reader_name = reader_name
        self._callback = callback or (lambda _: _)
        self._reader = None
        self._reader_observer = _CcidReaderObserver(self)
        self._card_observer = _CcidCardObserver(self)
        try:
            self._update(System.readers(), [])
        except EstablishContextException:
            pass  # No PC/SC context!

    def _update(self, added, removed):
        if self._reader in removed:  # Device removed
            self.reader = None
            self._set_status(CardStatus.NoCard)

        if self._reader is None:
            for reader in added:
                if self.reader_name in reader.name:
                    self.reader = reader
                    self._set_status(CardStatus.Present)
                    return

    @property
    def status(self):
        return self._status

    def _set_status(self, value):
        if self._status != value:
            self._status = value
            self.status_changed.emit(value)

    @property
    def reader(self):
        return self._reader

    @reader.setter
    def reader(self, value):
        self._reader = value
        self._callback(self, value)

    def open(self):
        if self._reader:
            conn = self._reader.createConnection()
            try:
                conn.connect()
                self._set_status(CardStatus.Present)
                return ScardDevice(conn)
            except SmartcardException:
                self._set_status(CardStatus.InUse)

    def __del__(self):
        self._reader_observer.delete()
        self._card_observer.delete()


def observe_reader(reader_name='Yubikey', callback=None):
    return CardWatcher(reader_name, callback)



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from __future__ import absolute_import

from PySide import QtGui, QtCore
from functools import partial
from os import getenv
from .utils import get_active_window, default_messages
import traceback


class _Messages(object):
    wait = 'Please wait...'


class _Event(QtCore.QEvent):
    EVENT_TYPE = QtCore.QEvent.Type(QtCore.QEvent.registerEventType())

    def __init__(self, callback):
        super(_Event, self).__init__(_Event.EVENT_TYPE)
        self._callback = callback

    def callback(self):
        self._callback()
        del self._callback


class Worker(QtCore.QObject):
    _work_signal = QtCore.Signal(tuple)
    _work_done_0 = QtCore.Signal()

    @default_messages(_Messages)
    def __init__(self, window, m):
        super(Worker, self).__init__()

        self.window = window

        self.busy = QtGui.QProgressDialog('', None, 0, 0, window)
        self.busy.setWindowTitle(m.wait)
        self.busy.setWindowModality(QtCore.Qt.WindowModal)
        self.busy.setMinimumDuration(0)
        self.busy.setWindowFlags(self.busy.windowFlags()
                                 ^ QtCore.Qt.WindowContextHelpButtonHint)
        self.busy.setAutoClose(True)

        self.work_thread = QtCore.QThread()
        self.moveToThread(self.work_thread)
        self.work_thread.start()

        self._work_signal.connect(self.work)
        self._work_done_0.connect(self.busy.reset)

    def post(self, title, fn, callback=None, return_errors=False):
        self.busy.setLabelText(title)
        self.busy.adjustPosition(get_active_window())
        self.busy.show()
        self.post_bg(fn, callback, return_errors)

    def post_bg(self, fn, callback=None, return_errors=False):
        if isinstance(fn, tuple):
            fn = partial(fn[0], *fn[1:])
        self._work_signal.emit((fn, callback, return_errors))

    def post_fg(self, fn):
        if isinstance(fn, tuple):
            fn = partial(fn[0], *fn[1:])
        event = _Event(fn)
        QtGui.QApplication.postEvent(self.window, event)

    @QtCore.Slot(tuple)
    def work(self, job):
        QtCore.QThread.msleep(10)  # Needed to yield
        (fn, callback, return_errors) = job
        try:
            result = fn()
        except Exception as e:
            result = e
            if getenv('DEBUG'):
                traceback.print_exc()
            if not return_errors:
                def callback(e): raise e
        if callback:
            event = _Event(partial(callback, result))
            QtGui.QApplication.postEvent(self.window, event)
        self._work_done_0.emit()



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from yubioath.yubicommon import qt
from .. import messages as m
from PySide import QtGui


class SetPasswordDialog(qt.Dialog):

    def __init__(self, parent):
        super(SetPasswordDialog, self).__init__(parent)

        self.setWindowTitle(m.set_pass)
        self._build_ui()

    def _build_ui(self):
        layout = QtGui.QFormLayout(self)

        self._new_pass = QtGui.QLineEdit()
        self._new_pass.setEchoMode(QtGui.QLineEdit.Password)
        layout.addRow(m.new_pass, self._new_pass)

        self._ver_pass = QtGui.QLineEdit()
        self._ver_pass.setEchoMode(QtGui.QLineEdit.Password)
        layout.addRow(m.ver_pass, self._ver_pass)

        self._remember = QtGui.QCheckBox(m.remember)
        layout.addRow(self._remember)

        btns = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok |
                                      QtGui.QDialogButtonBox.Cancel)
        btns.accepted.connect(self._save)
        btns.rejected.connect(self.reject)
        layout.addRow(btns)

    def _save(self):
        if not self._new_pass.text() == self._ver_pass.text():
            self._new_pass.setText('')
            self._ver_pass.setText('')
            self._new_pass.setFocus()
            QtGui.QMessageBox.warning(self, m.pass_mismatch,
                                      m.pass_mismatch_desc)
        else:
            self.accept()

    @property
    def password(self):
        return self._new_pass.text()

    @property
    def remember(self):
        return self._remember.isChecked()



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.


class CardError(Exception):

    def __init__(self, status, message=''):
        super(CardError, self).__init__('Card Error (%04x): %s' %
                                        (status, message))
        self.status = status


class DeviceLockedError(Exception):

    def __init__(self):
        super(DeviceLockedError, self).__init__('Device is locked!')


class NoSpaceError(Exception):

    def __init__(self):
        super(NoSpaceError, self).__init__('No space available on device.')


class InvalidSlotError(Exception):

    def __init__(self):
        super(InvalidSlotError, self).__init__(
            'The selected slot does not contain a valid OATH credential.')


class NeedsTouchError(Exception):

    def __init__(self):
        super(NeedsTouchError, self).__init__(
            'The selected slot needs touch to be used.')



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from PySide import QtGui, QtCore
from .. import messages as m
from ...core.standard import TYPE_TOTP, TYPE_HOTP
from yubioath.yubicommon.qt.utils import connect_once
from time import time


TIMELEFT_STYLE = """
QProgressBar {
  padding: 1px;
}
QProgressBar::chunk {
  background-color: #2196f3;
  margin: 0px;
  width: 1px;
}
"""


class TimeleftBar(QtGui.QProgressBar):
    expired = QtCore.Signal()

    def __init__(self):
        super(TimeleftBar, self).__init__()

        self.setStyleSheet(TIMELEFT_STYLE)
        self.setMaximumHeight(8)
        self.setInvertedAppearance(True)
        self.setRange(0, 30000)
        self.setValue(0)
        self.setTextVisible(False)

        self._timer = 0
        self._timeleft = 0

    def set_timeleft(self, millis):
        self._timeleft = max(0, millis)
        self.setValue(min(millis, self.maximum()))
        if self._timer == 0 and millis > 0:
            self._timer = self.startTimer(250)
        elif self._timer != 0 and millis <= 0:
            self.killTimer(self._timer)
            self._timer = 0

    def timerEvent(self, event):
        self.set_timeleft(max(0, self._timeleft - 250))
        if self._timeleft == 0:
            self.expired.emit()


class CodeMenu(QtGui.QMenu):

    def __init__(self, parent):
        super(CodeMenu, self).__init__(parent)
        self.entry = parent.entry

        self.addAction(m.action_delete).triggered.connect(self._delete)

    def _delete(self):
        res = QtGui.QMessageBox.warning(self, m.delete_title,
                                        m.delete_desc_1 % self.entry.cred.name,
                                        QtGui.QMessageBox.Ok,
                                        QtGui.QMessageBox.Cancel)
        if res == QtGui.QMessageBox.Ok:
            self.entry.delete()


class Code(QtGui.QWidget):

    def __init__(self, entry, timer, on_change):
        super(Code, self).__init__()
        self.entry = entry
        self.issuer, self.name = self._split_issuer_name()
        self._on_change = on_change
        self.entry.changed.connect(self._draw)
        self.timer = timer
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._menu)

        self._build_ui()

    def _build_ui(self):
        layout = QtGui.QHBoxLayout(self)
        labels = QtGui.QVBoxLayout()

        if self.issuer:
            self._issuer_lbl = QtGui.QLabel(self.issuer)
            labels.addWidget(self._issuer_lbl)

        self._code_lbl = QtGui.QLabel()
        labels.addWidget(self._code_lbl)

        self._name_lbl = QtGui.QLabel(self.name)
        labels.addWidget(self._name_lbl)

        layout.addLayout(labels)
        layout.addStretch()

        self._calc_btn = QtGui.QPushButton(QtGui.QIcon(':/calc.png'), None)
        self._calc_btn.clicked.connect(self._calc)
        layout.addWidget(self._calc_btn)
        self._calc_btn.setVisible(self.entry.manual)

        self._copy_btn = QtGui.QPushButton(QtGui.QIcon(':/copy.png'), None)
        self._copy_btn.clicked.connect(self._copy)
        layout.addWidget(self._copy_btn)

        self.timer.time_changed.connect(self._draw)

        self._draw()

    @property
    def expired(self):
        code = self.entry.code
        return code.timestamp + code.ttl <= self.timer.time

    def _draw(self):
        if self.expired:
            name_fmt = '<h2 style="color: gray;">%s</h2>'
        else:
            name_fmt = '<h2>%s</h2>'
        code = self.entry.code
        if self.entry.manual and self.entry.cred.oath_type != TYPE_HOTP:
            self._calc_btn.setEnabled(self.expired)
        self._code_lbl.setText(name_fmt % (code.code))
        self._copy_btn.setEnabled(bool(code.code))
        self._on_change()

    def _copy(self):
        QtCore.QCoreApplication.instance().clipboard().setText(
            self.entry.code.code)

    def _calc(self):
        if self.entry.manual:
            self._calc_btn.setDisabled(True)
        self.entry.calculate()
        if self.entry.cred.oath_type == TYPE_HOTP:
            QtCore.QTimer.singleShot(
                5000, lambda: self._calc_btn.setEnabled(True))

    def _menu(self, pos):
        CodeMenu(self).popup(self.mapToGlobal(pos))

    def _split_issuer_name(self):
        parts = self.entry.cred.name.split(':', 1)
        if len(parts) == 2:
            return parts
        return None, self.entry.cred.name

    def mouseDoubleClickEvent(self, event):
        if event.button() is QtCore.Qt.LeftButton:
            if (not self.entry.code.code or self.expired) and \
                    self.entry.manual:
                def copy_close():
                    self._copy()
                    self.window().close()
                connect_once(self.entry.changed, copy_close)
                self.entry.calculate()
            else:
                self._copy()  # TODO: Type code out with keyboard?
                self.window().close()
        event.accept()


class CodesList(QtGui.QWidget):

    def __init__(self, timer, credentials=[], on_change=None):
        super(CodesList, self).__init__()

        layout = QtGui.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        for cred in credentials:
            layout.addWidget(Code(cred, timer, on_change))
            line = QtGui.QFrame()
            line.setFrameShape(QtGui.QFrame.HLine)
            line.setFrameShadow(QtGui.QFrame.Sunken)
            layout.addWidget(line)

        if not credentials:
            no_creds = QtGui.QLabel(m.no_creds)
            no_creds.setAlignment(QtCore.Qt.AlignCenter)
            layout.addStretch()
            layout.addWidget(no_creds)
            layout.addStretch()

        layout.addStretch()


class CodesWidget(QtGui.QWidget):

    def __init__(self, controller):
        super(CodesWidget, self).__init__()

        self._controller = controller
        controller.refreshed.connect(self.refresh)
        controller.timer.time_changed.connect(self.refresh_timer)

        self._build_ui()
        self.refresh()
        self.refresh_timer()

    def _build_ui(self):
        layout = QtGui.QVBoxLayout(self)
        self._timeleft = TimeleftBar()
        layout.addWidget(self._timeleft)

        self._scroll_area = QtGui.QScrollArea()
        self._scroll_area.setWidgetResizable(True)
        self._scroll_area.setHorizontalScrollBarPolicy(
            QtCore.Qt.ScrollBarAlwaysOff)
        self._scroll_area.setVerticalScrollBarPolicy(
            QtCore.Qt.ScrollBarAsNeeded)
        self._scroll_area.setWidget(QtGui.QWidget())
        layout.addWidget(self._scroll_area)

    def refresh_timer(self, timestamp=None):
        if timestamp is None:
            timestamp = self._controller.timer.time
        if self._controller.has_expiring(timestamp):
            self._timeleft.set_timeleft(1000 * (timestamp + 30 - time()))
        else:
            self._timeleft.set_timeleft(0)

    def refresh(self):
        self._scroll_area.takeWidget().deleteLater()
        creds = self._controller.credentials
        self._scroll_area.setWidget(
            CodesList(self._controller.timer, creds or [], self.refresh_timer))
        w = self._scroll_area.widget().minimumSizeHint().width()
        w += self._scroll_area.verticalScrollBar().width()
        self._scroll_area.setMinimumWidth(w)



# Copyright (c) 2013 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

from PySide import QtGui
from .utils import *
from .classes import *
from .worker import *
from .settings import *
import sys
import traceback


# Font fixes for OSX
if sys.platform == 'darwin':
    from platform import mac_ver
    mac_version = tuple(int(x) for x in mac_ver()[0].split('.'))
    if (10, 9) <= mac_version < (10, 10):  # Mavericks
        QtGui.QFont.insertSubstitution('.Lucida Grande UI', 'Lucida Grande')
    if (10, 10) <= mac_version:  # Yosemite
        QtGui.QFont.insertSubstitution('.Helvetica Neue DeskInterface',
                                       'Helvetica Neue')


# Replace excepthook with one that releases the exception to prevent memory
# leaks:
def excepthook(typ, val, tback):
    try:
        traceback.print_exception(typ, val, tback)
        sys.exc_clear()
        del sys.last_value
        del sys.last_traceback
        del sys.last_type
    except:
        pass  # Ignore failure here, we're likely shutting down...
sys.excepthook = excepthook



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from yubioath.yubicommon import qt
from .. import messages as m
from PySide import QtGui

INDENT = 16


class SettingsDialog(qt.Dialog):

    def __init__(self, parent, settings):
        super(SettingsDialog, self).__init__(parent)
        self.settings = settings

        self.setWindowTitle(m.settings)
        self.accepted.connect(self._save)
        self._build_ui()
        self._reset()

    def _build_ui(self):
        layout = QtGui.QFormLayout(self)
        layout.addRow(self.section(m.ykstd_slots))

        # YubiKey slot 1
        self._slot1_enabled = QtGui.QCheckBox(m.enable_slot_1 % 1)
        self._slot1_enabled.setToolTip(m.tt_slot_enabled_1 % 1)
        layout.addRow(self._slot1_enabled)
        self._slot1_digits = QtGui.QComboBox()
        self._slot1_digits.addItems(['6', '8'])
        self._slot1_enabled.stateChanged.connect(self._slot1_digits.setEnabled)
        self._slot1_digits.setEnabled(False)
        self._slot1_digits.setToolTip(m.tt_num_digits)
        layout.addRow(m.n_digits, self._slot1_digits)
        layout.labelForField(self._slot1_digits).setIndent(INDENT)

        # YubiKey slot 2
        self._slot2_enabled = QtGui.QCheckBox(m.enable_slot_1 % 2)
        self._slot2_enabled.setToolTip(m.tt_slot_enabled_1 % 2)
        layout.addRow(self._slot2_enabled)
        self._slot2_digits = QtGui.QComboBox()
        self._slot2_digits.addItems(['6', '8'])
        self._slot2_enabled.stateChanged.connect(self._slot2_digits.setEnabled)
        self._slot2_digits.setEnabled(False)
        self._slot2_digits.setToolTip(m.tt_num_digits)
        layout.addRow(m.n_digits, self._slot2_digits)
        layout.labelForField(self._slot2_digits).setIndent(INDENT)

        layout.addRow(self.section(m.advanced))

        # Systray
        self._systray = QtGui.QCheckBox(m.enable_systray)
        self._systray.setToolTip(m.tt_systray)
        layout.addRow(self._systray)

        # Kill scdaemon
        self._kill_scdaemon = QtGui.QCheckBox(m.kill_scdaemon)
        self._kill_scdaemon.setToolTip(m.tt_kill_scdaemon)
        layout.addRow(self._kill_scdaemon)

        # Reader name
        self._reader_name = QtGui.QLineEdit()
        self._reader_name.setToolTip(m.tt_reader_name)
        layout.addRow(m.reader_name, self._reader_name)

        btns = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok |
                                      QtGui.QDialogButtonBox.Cancel)
        btns.accepted.connect(self.accept)
        btns.rejected.connect(self.reject)
        layout.addRow(btns)

    def _reset(self):
        slot1 = self.settings.get('slot1', 0)
        self._slot1_digits.setCurrentIndex(1 if slot1 == 8 else 0)
        self._slot1_enabled.setChecked(bool(slot1))

        slot2 = self.settings.get('slot2', 0)
        self._slot2_digits.setCurrentIndex(1 if slot2 == 8 else 0)
        self._slot2_enabled.setChecked(bool(slot2))

        self._systray.setChecked(self.settings.get('systray', False))
        self._kill_scdaemon.setChecked(
            self.settings.get('kill_scdaemon', False))

        self._reader_name.setText(self.settings.get('reader', 'Yubikey'))

    @property
    def slot1(self):
        return self._slot1_enabled.isChecked() \
            and int(self._slot1_digits.currentText())

    @property
    def slot2(self):
        return self._slot2_enabled.isChecked() \
            and int(self._slot2_digits.currentText())

    @property
    def systray(self):
        return self._systray.isChecked()

    @property
    def kill_scdaemon(self):
        return self._kill_scdaemon.isChecked()

    @property
    def reader_name(self):
        return self._reader_name.text()

    def _save(self):
        self.settings['slot1'] = self.slot1
        self.settings['slot2'] = self.slot2
        self.settings['systray'] = self.systray
        self.settings['kill_scdaemon'] = self.kill_scdaemon
        self.settings['reader'] = self.reader_name



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from ..cli.keystore import get_keystore as _get_keystore


def get_keystore():
    return Keystore(_get_keystore())


class Keystore(object):

    def __init__(self, store):
        self._store = store
        self._data = {}

    def get(self, id):
        return self._data.get(id) or self._store.get(id)

    def put(self, id, key, remember=False):
        if remember:
            self._store.put(id, key)
        else:
            self._data[id] = key

    def delete(self, id):
        if id in self._data:
            del self._data[id]
        self._store.delete(id)

    def forget(self):
        self._data = {}



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from PySide import QtGui
from .. import messages as m
import sys


class Systray(QtGui.QSystemTrayIcon):

    def __init__(self, parent):
        super(Systray, self).__init__(parent)

        self._reason = QtGui.QSystemTrayIcon.ActivationReason.Trigger

        # Require double-click on OSX since single click opens menu.
        if sys.platform == 'darwin':
            self._reason = QtGui.QSystemTrayIcon.ActivationReason.DoubleClick

        self.activated.connect(self._activated)
        self._build_menu()

    def _build_menu(self):
        menu = QtGui.QMenu()

        show_action = QtGui.QAction(m.action_show, menu)
        show_action.triggered.connect(self._show)
        menu.addAction(show_action)

        quit_action = QtGui.QAction(m.action_quit, menu)
        quit_action.triggered.connect(self.quit)
        menu.addAction(quit_action)

        self.setContextMenu(menu)

    def _activated(self, reason):
        if reason == self._reason:
            self._show()

    def _show(self):
        self.parent().window.show()
        self.parent().window.activateWindow()
        self.parent().window.raise_()

    def quit(self):
        self.hide()
        self.parent().window.close()



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from __future__ import print_function, division

from .exc import CardError, DeviceLockedError, NoSpaceError
from .utils import (der_read, der_pack, hmac_sha1, derive_key, get_random_bytes,
                    time_challenge, parse_truncated, format_code)
from yubioath.yubicommon.compat import int2byte, byte2int
import hashlib
import struct
import os

YKOATH_AID = b'\xa0\x00\x00\x05\x27\x21\x01\x01'
YKOATH_NO_SPACE = 0x6a84

INS_PUT = 0x01
INS_DELETE = 0x02
INS_SET_CODE = 0x03
INS_RESET = 0x04
INS_LIST = 0xa1
INS_CALCULATE = 0xa2
INS_VALIDATE = 0xa3
INS_CALC_ALL = 0xa4
INS_SEND_REMAINING = 0xa5

RESP_MORE_DATA = 0x61

TAG_NAME = 0x71
TAG_NAME_LIST = 0x72
TAG_KEY = 0x73
TAG_CHALLENGE = 0x74
TAG_RESPONSE = 0x75
TAG_T_RESPONSE = 0x76
TAG_NO_RESPONSE = 0x77
TAG_PROPERTY = 0x78
TAG_VERSION = 0x79
TAG_IMF = 0x7a
TAG_ALGO = 0x7b
TAG_TOUCH_RESPONSE = 0x7c

TYPE_MASK = 0xf0
TYPE_HOTP = 0x10
TYPE_TOTP = 0x20

ALG_MASK = 0x0f
ALG_SHA1 = 0x01
ALG_SHA256 = 0x02

PROP_ALWAYS_INC = 0x01
PROP_REQUIRE_TOUCH = 0x02

SCHEME_STANDARD = 0x00
SCHEME_STEAM = 0x01

STEAM_CHAR_TABLE = "23456789BCDFGHJKMNPQRTVWXY"


def format_code_steam(int_data, digits):
    chars = []
    for i in range(5):
        chars.append(STEAM_CHAR_TABLE[int_data % len(STEAM_CHAR_TABLE)])
        int_data //= len(STEAM_CHAR_TABLE)
    return ''.join(chars)


def ensure_unlocked(ykoath):
    if ykoath.locked:
        raise DeviceLockedError()


def format_truncated(t_resp, scheme=SCHEME_STANDARD):
    digits, data = byte2int(t_resp[0]), t_resp[1:]
    int_data = parse_truncated(data)
    if scheme == SCHEME_STANDARD:
        return format_code(int_data, digits)
    elif scheme == SCHEME_STEAM:
        return format_code_steam(int_data, digits)


def hmac_shorten_key(key, algo):
    if algo == ALG_SHA1:
        h = hashlib.sha1()
    elif algo == ALG_SHA256:
        h = hashlib.sha256()
    else:
        raise ValueError('Unsupported algorithm!')

    if len(key) > h.block_size:
        key = h.update(key).digest()

    return key


class Credential(object):
    """
    Reference to a credential.
    """

    def __init__(self, ykoath, oath_type, name, touch=False):
        self._ykoath = ykoath
        self.oath_type = oath_type
        self.name = name
        self.touch = touch

    def calculate(self, timestamp=None):
        return self._ykoath.calculate(self.name, self.oath_type, timestamp)

    def delete(self):
        self._ykoath.delete(self.name)

    def __repr__(self):
        return self.name


class _426Device(object):
    """
    The 4.2.0-4.2.6 firmwares have a known issue with credentials that require
    touch: If this action is performed within 2 seconds of a command resulting
    in a long response (over 54 bytes), the command will hang. A workaround is
    to send an invalid command (resulting in a short reply) prior to the
    "calculate" command.
    """

    def __init__(self, delegate):
        self._delegate = delegate
        self._long_resp = False

    def __getattr__(self, name):
        return getattr(self._delegate, name)

    def send_apdu(self, cl, ins, p1, p2, data):
        if ins == INS_CALCULATE and self._long_resp:
            self._delegate.send_apdu(0, 0, 0, 0, '')
            self._long_resp = False
        resp, status = self._delegate.send_apdu(cl, ins, p1, p2, data)
        self._long_resp = len(resp) > 52  # 52 bytes resp, 2 bytes status...
        return resp, status


class YubiOathCcid(object):

    """
    Device interface to a OATH-enabled YubiKey.
    """

    def __init__(self, device):
        self._device = device

        self._select()

        if (4, 2, 0) <= self.version <= (4, 2, 6):
            self._device = _426Device(device)

    def _send(self, ins, data='', p1=0, p2=0, expected=0x9000):
        resp, status = self._device.send_apdu(0, ins, p1, p2, data)
        while (status >> 8) == RESP_MORE_DATA:
            more, status = self._device.send_apdu(
                0, INS_SEND_REMAINING, 0, 0, '')
            resp += more

        if status == YKOATH_NO_SPACE:
            raise NoSpaceError()

        if expected != status:
            raise CardError(status)
        return resp

    def _select(self):
        resp = self._send(0xa4, YKOATH_AID, p1=0x04)
        self._version, resp = der_read(resp, TAG_VERSION)
        self._id, resp = der_read(resp, TAG_NAME)
        if len(resp) != 0:
            self._challenge, resp = der_read(resp, TAG_CHALLENGE)
        else:
            self._challenge = None

    @property
    def id(self):
        return self._id

    @property
    def version(self):
        return tuple(byte2int(d) for d in self._version)

    @property
    def locked(self):
        return self._challenge is not None

    def delete(self, name):
        data = der_pack(TAG_NAME, name.encode('utf8'))
        self._send(INS_DELETE, data)

    def calculate(self, name, oath_type, timestamp=None):
        challenge = time_challenge(timestamp) if oath_type == TYPE_TOTP else ''
        data = der_pack(TAG_NAME, name.encode('utf8'), TAG_CHALLENGE, challenge)
        resp = self._send(INS_CALCULATE, data)
        # Manual dynamic truncation required for Steam entries
        resp = der_read(resp, TAG_RESPONSE)[0]
        digits, resp = resp[0:1], resp[1:]
        offset = byte2int(resp[-1]) & 0xF
        resp = resp[offset:offset + 4]
        scheme = SCHEME_STEAM if name.startswith('Steam:') else SCHEME_STANDARD
        return format_truncated(digits + resp, scheme)

    def calculate_key(self, passphrase):
        return derive_key(self.id, passphrase)

    def unlock(self, key):
        if not self.locked:
            return

        response = hmac_sha1(key, self._challenge)
        challenge = get_random_bytes(8)
        verification = hmac_sha1(key, challenge)

        data = der_pack(TAG_RESPONSE, response, TAG_CHALLENGE, challenge)
        resp = self._send(INS_VALIDATE, data)
        resp = der_read(resp, TAG_RESPONSE)[0]
        if resp != verification:
            raise ValueError('Response did not match verification!')
        self._challenge = None

    def set_key(self, key=None):
        ensure_unlocked(self)
        if key:
            keydata = int2byte(TYPE_TOTP | ALG_SHA1) + key
            challenge = get_random_bytes(8)
            response = hmac_sha1(key, challenge)
            data = der_pack(TAG_KEY, keydata, TAG_CHALLENGE, challenge,
                            TAG_RESPONSE, response)
        else:
            data = der_pack(TAG_KEY, b'')
        self._send(INS_SET_CODE, data)

    def reset(self):
        self._send(INS_RESET, p1=0xde, p2=0xad)
        self._challenge = None

    def list(self):
        ensure_unlocked(self)
        resp = self._send(INS_LIST)
        items = []
        while resp:
            data, resp = der_read(resp, TAG_NAME_LIST)
            items.append(Credential(
                self,
                TYPE_MASK & byte2int(data[0]),
                data[1:],
                None
            ))
        return items

    def calculate_all(self, timestamp=None):
        ensure_unlocked(self)
        data = der_pack(TAG_CHALLENGE, time_challenge(timestamp))
        resp = self._send(INS_CALC_ALL, data, p2=1)
        results = []
        while resp:
            name, resp = der_read(resp, TAG_NAME)
            name = name.decode('utf8')
            tag, value, resp = der_read(resp)
            if name.startswith('_hidden:') and 'YKOATH_SHOW_HIDDEN' not in os.environ:
                pass  # Ignore hidden credentials.
            elif tag == TAG_T_RESPONSE:
                # Steam credentials need to be recalculated
                # to skip full truncation done by Yubikey 4.
                code = self.calculate(name, TYPE_TOTP) \
                        if name.startswith('Steam:') \
                        else format_truncated(value, SCHEME_STANDARD)
                results.append((
                    Credential(self, TYPE_TOTP, name, False),
                    code
                ))
            elif tag == TAG_TOUCH_RESPONSE:
                results.append((
                    Credential(self, TYPE_TOTP, name, True),
                    None
                ))
            elif tag == TAG_NO_RESPONSE:
                results.append((Credential(self, TYPE_HOTP, name), None))
            else:
                print("Unsupported tag: %02x" % tag)
        results.sort(key=lambda a: a[0].name.lower())
        return results

    def put(self, name, key, oath_type=TYPE_TOTP, algo=ALG_SHA1, digits=6,
            imf=0, always_increasing=False, require_touch=False):
        ensure_unlocked(self)
        key = hmac_shorten_key(key, algo)
        keydata = int2byte(oath_type | algo) + int2byte(digits) + key
        data = der_pack(TAG_NAME, name.encode('utf8'), TAG_KEY, keydata)
        properties = 0
        if always_increasing:
            properties |= PROP_ALWAYS_INC
        if require_touch:
            if self.version < (4, 2, 6):
                raise Exception("Touch-required not supported on this key")
            properties |= PROP_REQUIRE_TOUCH
        if properties:
            data += int2byte(TAG_PROPERTY) + int2byte(properties)
        if imf > 0:
            data += der_pack(TAG_IMF, struct.pack('>I', imf))
        self._send(INS_PUT, data)
        return Credential(self, oath_type, name)



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

__version__ = "3.0.0"



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from __future__ import print_function

from smartcard import System
from smartcard.Exceptions import SmartcardException

from yubioath.yubicommon.compat import byte2int, int2byte


class ScardDevice(object):

    """
    Pyscard based backend.
    """

    def __init__(self, connection):
        self._conn = connection

    def send_apdu(self, cl, ins, p1, p2, data):
        header = [cl, ins, p1, p2, len(data)]
        # from binascii import b2a_hex
        # print("SEND:", b2a_hex(''.join(map(int2byte, header)) + data))
        resp, sw1, sw2 = self._conn.transmit(header + [byte2int(b) for b in data])
        # print("RECV:", b2a_hex(b''.join(map(int2byte, resp + [sw1, sw2]))))
        return b''.join(int2byte(i) for i in resp), sw1 << 8 | sw2

    def close(self):
        self._conn.disconnect()

    def __del__(self):
        self.close()


def open_scard(name='Yubikey'):
    name = name.lower()
    for reader in System.readers():
        if name in reader.name.lower():
            conn = reader.createConnection()
            try:
                conn.connect()
                return ScardDevice(conn)
            except SmartcardException:
                pass



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.


from .standard import YubiOathCcid
from .legacy_ccid import LegacyOathCcid
from .exc import CardError, InvalidSlotError, NeedsTouchError
import time
import sys
try:
    from .legacy_otp import open_otp, LegacyOathOtp, LegacyCredential
except ImportError:
    sys.stderr.write('libykpers not found!\n')
    open_otp = None


class Controller(object):

    @property
    def otp_supported(self):
        return bool(open_otp)

    def open_otp(self):
        otp_dev = open_otp()
        if otp_dev:
            return LegacyOathOtp(otp_dev)

    def _prompt_touch(self):
        pass

    def _end_prompt_touch(self):
        pass

    def unlock(self, std):
        raise ValueError('Password required')

    def read_slot_ccid(self, std, slot, digits, timestamp=None):
        cred = LegacyCredential(std, slot, digits)
        try:
            return (cred, cred.calculate(timestamp))
        except InvalidSlotError:
            return (cred, 'INVALID')

    def read_slot_otp(self, cred, timestamp=None, use_touch=False):
        if cred.touch:
            if not use_touch:
                raise NeedsTouchError()
            self._prompt_touch()

        try:
            return (cred, cred.calculate(timestamp))
        except InvalidSlotError:
            return (cred, 'INVALID')
        except NeedsTouchError:
            if use_touch:
                try:
                    time.sleep(0.1)  # Give the key a little time...
                    self._prompt_touch()
                    start = time.time()
                    return (cred, cred.calculate(timestamp))
                except InvalidSlotError:
                    error = 'INVALID' if time.time() - start < 1 else 'TIMEOUT'
                    return (cred, error)
            return (cred, None)
        finally:
            if cred.touch:
                self._end_prompt_touch()

    def read_creds(self, ccid_dev, slot1, slot2, timestamp, mayblock=True):
        results = []
        key_found = False
        do_legacy = mayblock and bool(slot1 or slot2)
        legacy_creds = [None, None]

        if ccid_dev:
            try:
                std = YubiOathCcid(ccid_dev)
                key_found = True
                if std.locked:
                    self.unlock(std)
                results.extend(std.calculate_all(timestamp))
            except CardError:
                pass  # No applet?

            if do_legacy:
                try:
                    legacy = LegacyOathCcid(ccid_dev)
                    for (slot, digits) in [(0, slot1), (1, slot2)]:
                        if digits:
                            try:
                                legacy_creds[slot] = self.read_slot_ccid(
                                    legacy, slot+1, digits, timestamp)
                            except NeedsTouchError:
                                pass  # Handled over OTP instead
                except CardError:
                    pass  # No applet?

        if self.otp_supported and ((slot1 and not legacy_creds[0])
                                   or (slot2 and not legacy_creds[1])):
            if ccid_dev:
                ccid_dev.close()
            legacy = self.open_otp()
            if legacy:
                key_found = True
                if not legacy_creds[0] and slot1:
                    legacy_creds[0] = self.read_slot_otp(
                        LegacyCredential(legacy, 1, slot1), timestamp, True)
                if not legacy_creds[1] and slot2:
                    legacy_creds[1] = self.read_slot_otp(
                        LegacyCredential(legacy, 2, slot2), timestamp, True)
                del legacy._device

        if not key_found:
            return None

        # Add legacy slots first.
        if legacy_creds[1]:
            results.insert(0, legacy_creds[1])
        if legacy_creds[0]:
            results.insert(0, legacy_creds[0])

        return results

    def set_password(self, dev, password):
        if dev.locked:
            self.unlock(dev)
        key = dev.calculate_key(password)
        dev.set_key(key)
        return key

    def add_cred(self, dev, *args, **kwargs):
        if dev.locked:
            self.unlock(dev)
        dev.put(*args, **kwargs)

    def add_cred_legacy(self, *args, **kwargs):
        legacy = self.open_otp()
        if not legacy:
            raise Exception('No YubiKey found!')
        legacy.put(*args, **kwargs)

    def delete_cred(self, dev, name):
        if dev.locked:
            self.unlock(dev)
        dev.delete(name)

    def delete_cred_legacy(self, slot):
        legacy = self.open_otp()
        if not legacy:
            raise Exception('No YubiKey found!')
        legacy.delete(slot)

    def reset_device(self, dev):
        dev.reset()



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from ..core.standard import (YubiOathCcid, TYPE_TOTP, TYPE_HOTP,
                             Credential as StdCredential)
from ..core.legacy_otp import LegacyCredential
from ..core.controller import Controller
from ..core.exc import CardError, DeviceLockedError
from .ccid import CardStatus
from yubioath.yubicommon.qt.utils import is_minimized
from .view.get_password import GetPasswordDialog
from .keystore import get_keystore
from . import messages as m
from yubioath.core.utils import ccid_supported_but_disabled
from yubioath.yubicommon.qt import get_active_window, MutexLocker
from PySide import QtCore, QtGui
from time import time
from collections import namedtuple

import sys
if sys.platform == 'win32':  # Windows has issues with the high level API.
    from .ccid_poll import observe_reader
else:
    from .ccid import observe_reader


Code = namedtuple('Code', 'code timestamp ttl')
UNINITIALIZED = Code('', 0, 0)

TIME_PERIOD = 30
INF = float('inf')


class CredEntry(QtCore.QObject):
    changed = QtCore.Signal()

    def __init__(self, cred, controller):
        super(CredEntry, self).__init__()
        self.cred = cred
        self._controller = controller
        self._code = Code('', 0, 0)

    @property
    def code(self):
        return self._code

    @code.setter
    def code(self, value):
        self._code = value
        self.changed.emit()

    @property
    def manual(self):
        return self.cred.touch or self.cred.oath_type == TYPE_HOTP

    def calculate(self):
        dialog = QtGui.QMessageBox(get_active_window())
        dialog.setWindowTitle(m.touch_title)
        dialog.setStandardButtons(QtGui.QMessageBox.NoButton)
        dialog.setIcon(QtGui.QMessageBox.Information)
        dialog.setText(m.touch_desc)
        timer = None

        def cb(code):
            if timer:
                timer.stop()
            dialog.accept()
            if isinstance(code, Exception):
                QtGui.QMessageBox.warning(get_active_window(), m.error,
                                          code.message)
            else:
                self.code = code
        self._controller._app.worker.post_bg((self._controller._calculate_cred,
                                              self.cred), cb)
        if self.cred.touch:
            dialog.exec_()
        elif self.cred.oath_type == TYPE_HOTP:
            # HOTP might require touch, we don't know. Assume yes after 500ms.
            timer = QtCore.QTimer(self)
            timer.setSingleShot(True)
            timer.timeout.connect(dialog.exec_)
            timer.start(500)

    def delete(self):
        if self.cred.name in ['YubiKey slot 1', 'YubiKey slot 2']:
            self._controller.delete_cred_legacy(int(self.cred.name[-1]))
        else:
            self._controller.delete_cred(self.cred.name)


Capabilities = namedtuple('Capabilities', 'ccid otp version')


def names(creds):
    return set(c.cred.name for c in creds)


class Timer(QtCore.QObject):
    time_changed = QtCore.Signal(int)

    def __init__(self, interval):
        super(Timer, self).__init__()
        self._interval = interval

        now = time()
        rem = now % interval
        self._time = int(now - rem)
        QtCore.QTimer.singleShot((self._interval - rem) * 1000, self._tick)

    def _tick(self):
        self._time += self._interval
        self.time_changed.emit(self._time)
        next_time = self._time + self._interval
        QtCore.QTimer.singleShot((next_time - time()) * 1000, self._tick)

    @property
    def time(self):
        return self._time


class GuiController(QtCore.QObject, Controller):
    refreshed = QtCore.Signal()
    ccid_disabled = QtCore.Signal()

    def __init__(self, app, settings):
        super(GuiController, self).__init__()
        self._app = app
        self._settings = settings
        self._needs_read = False
        self._reader = None
        self._creds = None
        self._lock = QtCore.QMutex()
        self._keystore = get_keystore()
        self._current_device_has_ccid_disabled = False
        self.timer = Timer(TIME_PERIOD)

        self.watcher = observe_reader(self.reader_name, self._on_reader)

        self.startTimer(3000)
        self.timer.time_changed.connect(self.refresh_codes)

    def settings_changed(self):
        self.watcher.reader_name = self.reader_name
        self.refresh_codes()

    @property
    def reader_name(self):
        return self._settings.get('reader', 'Yubikey')

    @property
    def slot1(self):
        return self._settings.get('slot1', 0)

    @property
    def slot2(self):
        return self._settings.get('slot2', 0)

    @property
    def mute_ccid_disabled_warning(self):
        return self._settings.get('mute_ccid_disabled_warning', 0)

    @mute_ccid_disabled_warning.setter
    def mute_ccid_disabled_warning(self, value):
        self._settings['mute_ccid_disabled_warning'] = value

    def unlock(self, dev):
        if dev.locked:
            key = self._keystore.get(dev.id)
            if not key:
                self._app.worker.post_fg((self._init_dev, dev))
                return False
            dev.unlock(key)
        return True

    def grab_lock(self, lock=None, try_lock=False):
        return lock or MutexLocker(self._lock, False).lock(try_lock)

    @property
    def otp_enabled(self):
        return self.otp_supported and bool(self.slot1 or self.slot2)

    @property
    def credentials(self):
        return self._creds

    def has_expiring(self, timestamp):
        for c in self._creds or []:
            if c.code.timestamp >= timestamp and c.code.ttl < INF:
                return True
        return False

    def get_capabilities(self):
        assert self.grab_lock()
        ccid_dev = self.watcher.open()
        if ccid_dev:
            dev = YubiOathCcid(ccid_dev)
            return Capabilities(True, None, dev.version)
        legacy = self.open_otp()
        if legacy:
            return Capabilities(None, legacy.slot_status(), (0, 0, 0))
        return Capabilities(None, None, (0, 0, 0))
    
    def get_entry_names(self):
        return names(self._creds)

    def _on_reader(self, watcher, reader, lock=None):
        if reader:
            if self._reader is None:
                self._reader = reader
                self._creds = []
                if is_minimized(self._app.window):
                    self._needs_read = True
                else:
                    ccid_dev = watcher.open()
                    if ccid_dev:
                        dev = YubiOathCcid(ccid_dev)
                        self._app.worker.post_fg((self._init_dev, dev))
                    else:
                        self._needs_read = True
            elif self._needs_read:
                self.refresh_codes(self.timer.time)
        else:
            self._reader = None
            self._creds = None
            self.refreshed.emit()

    def _init_dev(self, dev):
        lock = self.grab_lock()
        while dev.locked:
            if self._keystore.get(dev.id) is None:
                dialog = GetPasswordDialog(get_active_window())
                if dialog.exec_():
                    self._keystore.put(dev.id,
                                       dev.calculate_key(dialog.password),
                                       dialog.remember)
                else:
                    return
            try:
                dev.unlock(self._keystore.get(dev.id))
            except CardError:
                self._keystore.delete(dev.id)
        self.refresh_codes(self.timer.time, lock)

    def _await(self):
        self._creds = None

    def wrap_credential(self, tup):
        (cred, code) = tup
        entry = CredEntry(cred, self)
        if code and code not in ['INVALID', 'TIMEOUT']:
            entry.code = Code(code, self.timer.time, TIME_PERIOD)

        return entry

    def _set_creds(self, creds):
        if creds:
            creds = [self.wrap_credential(c) for c in creds]
            if self._creds and names(creds) == names(self._creds):
                entry_map = dict((c.cred.name, c) for c in creds)
                for entry in self._creds:
                    cred = entry.cred
                    code = entry_map[cred.name].code
                    if code.code:
                        entry.code = code
                    elif cred.oath_type != entry_map[cred.name].cred.oath_type:
                        break
                else:
                    return
            elif self._reader and self._needs_read and self._creds:
                return
        self._creds = creds
        self.refreshed.emit()

    def _calculate_cred(self, cred):
        assert self.grab_lock()
        now = time()
        timestamp = self.timer.time
        if timestamp + TIME_PERIOD - now < 10:
            timestamp += TIME_PERIOD
        ttl = TIME_PERIOD
        if cred.oath_type == TYPE_HOTP:
            ttl = INF

        if cred.name in ['YubiKey slot 1', 'YubiKey slot 2']:
            legacy = self.open_otp()
            if not legacy:
                raise ValueError('YubiKey removed!')

            try:
                cred._legacy = legacy
                cred, code = super(GuiController, self).read_slot_otp(
                    cred, timestamp, True)
            finally:
                cred._legacy = None  # Release the handle.
            return Code(code, timestamp, TIME_PERIOD)

        ccid_dev = self.watcher.open()
        if not ccid_dev:
            if self.watcher.status != CardStatus.Present:
                self._set_creds(None)
            return
        dev = YubiOathCcid(ccid_dev)
        if self.unlock(dev):
            return Code(dev.calculate(cred.name, cred.oath_type, timestamp),
                        timestamp, ttl)

    def read_slot_otp(self, cred, timestamp=None, use_touch=False):
        return super(GuiController, self).read_slot_otp(cred, timestamp, False)

    def refresh_codes(self, timestamp=None, lock=None):
        if not self._reader and self.watcher.reader:
            return self._on_reader(self.watcher, self.watcher.reader, lock)
        elif is_minimized(self._app.window):
            self._needs_read = True
            return
        lock = self.grab_lock(lock, True)
        if not lock:
            return
        device = self.watcher.open()
        self._needs_read = bool(self._reader and device is None)
        timestamp = timestamp or self.timer.time
        try:
            creds = self.read_creds(device, self.slot1, self.slot2, timestamp,
                                    False)
        except DeviceLockedError:
            creds = []
        self._set_creds(creds)

    def timerEvent(self, event):
        if not is_minimized(self._app.window):
            timestamp = self.timer.time
            if self._reader and self._needs_read:
                self._app.worker.post_bg(self.refresh_codes)
            elif self._reader is None:
                if self.otp_enabled:
                    def refresh_otp():
                        lock = self.grab_lock(try_lock=True)
                        if lock:
                            read = self.read_creds(None, self.slot1, self.slot2,
                                                   timestamp, False)
                            self._set_creds(read)
                    self._app.worker.post_bg(refresh_otp)
                else:
                    if ccid_supported_but_disabled():
                        if not self._current_device_has_ccid_disabled:
                            self.ccid_disabled.emit()
                        self._current_device_has_ccid_disabled = True
                        event.accept()
                        return
            self._current_device_has_ccid_disabled = False
        event.accept()

    def add_cred(self, *args, **kwargs):
        lock = self.grab_lock()
        ccid_dev = self.watcher.open()
        if ccid_dev:
            dev = YubiOathCcid(ccid_dev)
            if self.unlock(dev):
                super(GuiController, self).add_cred(dev, *args, **kwargs)
                self._creds = None
                self.refresh_codes(lock=lock)

    def add_cred_legacy(self, *args, **kwargs):
        lock = self.grab_lock()
        super(GuiController, self).add_cred_legacy(*args, **kwargs)
        self._creds = None
        self.refresh_codes(lock=lock)

    def delete_cred(self, name):
        lock = self.grab_lock()
        ccid_dev = self.watcher.open()
        if ccid_dev:
            dev = YubiOathCcid(ccid_dev)
            if self.unlock(dev):
                super(GuiController, self).delete_cred(dev, name)
                self._creds = None
                self.refresh_codes(lock=lock)

    def delete_cred_legacy(self, *args, **kwargs):
        lock = self.grab_lock()
        super(GuiController, self).delete_cred_legacy(*args, **kwargs)
        self._creds = None
        self.refresh_codes(lock=lock)

    def set_password(self, password, remember=False):
        assert self.grab_lock()
        ccid_dev = self.watcher.open()
        if ccid_dev:
            dev = YubiOathCcid(ccid_dev)
            if self.unlock(dev):
                key = super(GuiController, self).set_password(dev, password)
                self._keystore.put(dev.id, key, remember)

    def forget_passwords(self):
        self._keystore.forget()



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from yubioath.yubicommon import qt
from ...core.standard import TYPE_TOTP, TYPE_HOTP
from ...core.utils import parse_uri
from .. import messages as m
from ..qrparse import parse_qr_codes
from ..qrdecode import decode_qr_data
from PySide import QtGui, QtCore
from base64 import b32decode
import re

NAME_VALIDATOR = QtGui.QRegExpValidator(QtCore.QRegExp(r'.{3,}'))


class B32Validator(QtGui.QValidator):

    def __init__(self, parent=None):
        super(B32Validator, self).__init__(parent)
        self.partial = re.compile(r'^[ a-z2-7]+$', re.IGNORECASE)

    def fixup(self, value):
        unpadded = value.upper().rstrip('=').replace(' ', '')
        return b32decode(unpadded + '=' * (-len(unpadded) % 8))

    def validate(self, value, pos):
        try:
            self.fixup(value)
            return QtGui.QValidator.Acceptable
        except:
            if self.partial.match(value):
                return QtGui.QValidator.Intermediate
        return QtGui.QValidator.Invalid


class AddCredDialog(qt.Dialog):

    def __init__(self, worker, version, existing_entry_names, parent=None):
        super(AddCredDialog, self).__init__(parent)

        self._worker = worker
        self._version = version
        self.setWindowTitle(m.add_cred)
        self._existing_entry_names = existing_entry_names
        self._build_ui()

    def _build_ui(self):
        layout = QtGui.QFormLayout(self)

        self._qr_btn = QtGui.QPushButton(QtGui.QIcon(':/qr.png'), m.qr_scan)
        self._qr_btn.clicked.connect(self._scan_qr)
        layout.addRow(self._qr_btn)

        self._cred_name = QtGui.QLineEdit()
        self._cred_name.setValidator(NAME_VALIDATOR)
        layout.addRow(m.cred_name, self._cred_name)

        self._cred_key = QtGui.QLineEdit()
        self._cred_key.setValidator(B32Validator())
        layout.addRow(m.cred_key, self._cred_key)

        layout.addRow(QtGui.QLabel(m.cred_type))
        self._cred_type = QtGui.QButtonGroup(self)
        self._cred_totp = QtGui.QRadioButton(m.cred_totp)
        self._cred_totp.setProperty('value', TYPE_TOTP)
        self._cred_type.addButton(self._cred_totp)
        layout.addRow(self._cred_totp)
        self._cred_hotp = QtGui.QRadioButton(m.cred_hotp)
        self._cred_hotp.setProperty('value', TYPE_HOTP)
        self._cred_type.addButton(self._cred_hotp)
        layout.addRow(self._cred_hotp)
        self._cred_totp.setChecked(True)

        self._n_digits = QtGui.QComboBox()
        self._n_digits.addItems(['6', '8'])
        layout.addRow(m.n_digits, self._n_digits)

        self._require_touch = QtGui.QCheckBox(m.require_touch)
        # Touch-required support not available before 4.2.6
        if self._version >= (4, 2, 6):
            layout.addRow(self._require_touch)

        btns = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok |
                                      QtGui.QDialogButtonBox.Cancel)
        btns.accepted.connect(self._save)
        btns.rejected.connect(self.reject)
        layout.addRow(btns)

    def _do_scan_qr(self, qimage):
        for qr in parse_qr_codes(qimage):
            try:
                data = decode_qr_data(qr)
                if data.startswith('otpauth://'):
                    return parse_uri(data)
            except:
                pass
        return None

    def _scan_qr(self):
        winId = QtGui.QApplication.desktop().winId()
        qimage = QtGui.QPixmap.grabWindow(winId).toImage()
        self._worker.post(m.qr_scanning, (self._do_scan_qr, qimage),
                          self._handle_qr)

    def _handle_qr(self, parsed):
        if parsed:
            self._cred_name.setText(parsed['name'])
            self._cred_key.setText(parsed['secret'])
            n_digits = parsed.get('digits', '6')
            self._n_digits.setCurrentIndex(0 if n_digits == '6' else 1)
            if parsed['type'] == 'totp':
                self._cred_totp.setChecked(True)
            else:
                self._cred_hotp.setChecked(True)
        else:
            QtGui.QMessageBox.warning(self, m.qr_not_found, m.qr_not_found_desc)
    
    def _entry_exists(self):
        return self._cred_name.text() in self._existing_entry_names
    
    def _confirm_overwrite(self):
        return QtGui.QMessageBox.question(self, m.overwrite_entry, m.overwrite_entry_desc,
                QtGui.QMessageBox.Yes | QtGui.QMessageBox.Yes,
                QtGui.QMessageBox.No) == QtGui.QMessageBox.Yes

    def _save(self):
        if not self._cred_name.hasAcceptableInput():
            QtGui.QMessageBox.warning(self, m.invalid_name, m.invalid_name_desc)
            self._cred_name.selectAll()
        elif not self._cred_key.hasAcceptableInput():
            QtGui.QMessageBox.warning(self, m.invalid_key, m.invalid_key_desc)
            self._cred_key.selectAll()
        elif self._entry_exists() and not self._confirm_overwrite():
            self._cred_key.selectAll()
        else:
            self.accept()
    


    @property
    def name(self):
        return self._cred_name.text()

    @property
    def key(self):
        return self._cred_key.validator().fixup(self._cred_key.text())

    @property
    def oath_type(self):
        return self._cred_type.checkedButton().property('value')

    @property
    def n_digits(self):
        return int(self._n_digits.currentText())

    @property
    def require_touch(self):
        return self._require_touch.isChecked()



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.


from yubioath.yubicommon.setup import setup
from yubioath.yubicommon.setup.qt import qt_resources
from yubioath.yubicommon.setup.exe import executable

setup(
    name='yubioath-desktop',
    long_name='Yubico Authenticator',
    author='Dain Nilsson',
    author_email='dain@yubico.com',
    maintainer='Yubico Open Source Maintainers',
    maintainer_email='ossmaint@yubico.com',
    url='https://github.com/Yubico/yubioath-desktop',
    license='GPLv3+',
    description='Graphical interface for displaying OATH codes with a Yubikey',
    entry_points={
        'console_scripts': ['yubioath=yubioath.cli.__main__:main'],
        'gui_scripts': ['yubioath-gui=yubioath.gui.__main__:main']
    },
    setup_requires=[],
    yc_requires=['ctypes', 'qt'],
    yc_requires_exclude=['PySide'],
    install_requires=['pyscard', 'pycrypto'],
    extras_require={
        'cli': ['click'],
        'gui': ['PySide']
    },
    cmdclass={
        'executable': executable,
        'qt_resources': qt_resources('yubioath.gui')
    },
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Environment :: X11 Applications :: Qt',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Security :: Cryptography',
        'Topic :: Utilities'
    ]
)



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from yubioath.yubicommon import qt
from .add_cred import B32Validator
from .. import messages as m
from ..qrparse import parse_qr_codes
from ..qrdecode import decode_qr_data
from ...core.utils import parse_uri
from PySide import QtGui
from base64 import b32decode


class AddCredDialog(qt.Dialog):

    def __init__(self, worker, otp_slots=(0, 0), url=None, parent=None):
        super(AddCredDialog, self).__init__(parent)

        self.setWindowTitle(m.add_cred)
        self._worker = worker
        self._slot_status = otp_slots
        self._build_ui()

    def _build_ui(self):
        layout = QtGui.QFormLayout(self)

        self._qr_btn = QtGui.QPushButton(QtGui.QIcon(':/qr.png'), m.qr_scan)
        self._qr_btn.clicked.connect(self._scan_qr)
        layout.addRow(self._qr_btn)

        self._cred_key = QtGui.QLineEdit()
        self._cred_key.setValidator(B32Validator())
        layout.addRow(m.cred_key, self._cred_key)

        layout.addRow(QtGui.QLabel(m.slot))
        self._slot = QtGui.QButtonGroup(self)
        slot1_status = m.in_use if self._slot_status[0] else m.free
        self._slot_1 = QtGui.QRadioButton(m.slot_2 % (1, slot1_status))
        self._slot_1.setProperty('value', 1)
        self._slot.addButton(self._slot_1)
        layout.addRow(self._slot_1)
        slot2_status = m.in_use if self._slot_status[1] else m.free
        self._slot_2 = QtGui.QRadioButton(m.slot_2 % (2, slot2_status))
        self._slot_2.setProperty('value', 2)
        self._slot.addButton(self._slot_2)
        layout.addRow(self._slot_2)

        self._touch = QtGui.QCheckBox(m.require_touch)
        layout.addRow(self._touch)

        self._n_digits = QtGui.QComboBox()
        self._n_digits.addItems(['6', '8'])
        layout.addRow(m.n_digits, self._n_digits)

        btns = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok |
                                      QtGui.QDialogButtonBox.Cancel)
        btns.accepted.connect(self._save)
        btns.rejected.connect(self.reject)
        layout.addRow(btns)

    def _save(self):
        if not self._cred_key.hasAcceptableInput():
            QtGui.QMessageBox.warning(self, m.invalid_key, m.invalid_key_desc)
            self._cred_key.selectAll()
        elif not self._slot.checkedButton():
            QtGui.QMessageBox.warning(self, m.no_slot, m.no_slot_desc)
        elif self._slot_status[self.slot - 1] \
            and QtGui.QMessageBox.Ok != QtGui.QMessageBox.warning(
                self, m.overwrite_slot, m.overwrite_slot_desc_1 % self.slot,
                QtGui.QMessageBox.Ok | QtGui.QMessageBox.Cancel):
            return
        else:
            self.accept()

    def _do_scan_qr(self, qimage):
        for qr in parse_qr_codes(qimage):
            try:
                data = decode_qr_data(qr)
                if data.startswith('otpauth://'):
                    return parse_uri(data)
            except:
                pass
        return None

    def _scan_qr(self):
        winId = QtGui.QApplication.desktop().winId()
        qimage = QtGui.QPixmap.grabWindow(winId).toImage()
        self._worker.post(m.qr_scanning, (self._do_scan_qr, qimage),
                          self._handle_qr)

    def _handle_qr(self, parsed):
        if parsed:
            if parsed['type'] != 'totp':
                QtGui.QMessageBox.warning(self, m.qr_not_supported,
                                          m.qr_not_supported_desc)
            else:
                self._cred_key.setText(parsed['secret'])
                n_digits = parsed.get('digits', '6')
                self._n_digits.setCurrentIndex(0 if n_digits == '6' else 1)
        else:
            QtGui.QMessageBox.warning(self, m.qr_not_found, m.qr_not_found_desc)


    @property
    def key(self):
        unpadded = self._cred_key.text().upper()
        return b32decode(unpadded + '=' * (-len(unpadded) % 8))

    @property
    def slot(self):
        return self._slot.checkedButton().property('value')

    @property
    def touch(self):
        return self._touch.isChecked()

    @property
    def n_digits(self):
        return int(self._n_digits.currentText())



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Additional permission under GNU GPL version 3 section 7
#
# If you modify this program, or any covered work, by linking or
# combining it with the OpenSSL project's OpenSSL library (or a
# modified version of that library), containing parts covered by the
# terms of the OpenSSL or SSLeay licenses, We grant you additional
# permission to convey the resulting work. Corresponding Source for a
# non-source form of such a combination shall include the source code
# for the parts of OpenSSL used as well as that of the covered work.

from __future__ import absolute_import

from PySide import QtCore
from collections import MutableMapping

__all__ = ['Settings', 'PySettings', 'convert_to']


def convert_to(value, target_type):
    if target_type is list:
        return [] if value is None else [value]
    if target_type is int:
        return 0 if value in ['', 'false', 'False'] else int(value)
    if target_type is float:
        return float(value)
    if target_type is bool:
        return value not in ['', 'false', 'False']
    return value


class SettingsGroup(object):

    def __init__(self, settings, mutex, group):
        self._settings = settings
        self._mutex = mutex
        self._group = group

    def __getattr__(self, method_name):
        if hasattr(self._settings, method_name):
            fn = getattr(self._settings, method_name)

            def wrapped(*args, **kwargs):
                try:
                    self._mutex.lock()
                    self._settings.beginGroup(self._group)
                    return fn(*args, **kwargs)
                finally:
                    self._settings.endGroup()
                    self._mutex.unlock()
            return wrapped

    def rename(self, new_name):
        data = dict((key, self.value(key)) for key in self.childKeys())
        self.remove('')
        self._group = new_name
        for k, v in data.items():
            self.setValue(k, v)

    def __repr__(self):
        return 'Group(%s)' % self._group


class Settings(QtCore.QObject):

    def __init__(self, q_settings, wrap=True):
        super(Settings, self).__init__()
        self._mutex = QtCore.QMutex(QtCore.QMutex.Recursive)
        self._wrap = wrap
        self._q_settings = q_settings

    def get_group(self, group):
        g = SettingsGroup(self._q_settings, self._mutex, group)
        if self._wrap:
            g = PySettings(g)
        return g

    @staticmethod
    def wrap(*args, **kwargs):
        return Settings(QtCore.QSettings(*args, **kwargs))


class PySettings(MutableMapping):

    def __init__(self, settings):
        self._settings = settings

    def __getattr__(self, method_name):
        return getattr(self._settings, method_name)

    def get(self, key, default=None):
        val = self._settings.value(key, default)
        if not isinstance(val, type(default)):
            val = convert_to(val, type(default))
        return val

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self._settings.setValue(key, value)

    def __delitem__(self, key):
        self._settings.remove(key)

    def __iter__(self):
        for key in list(self.keys()):
            yield key

    def __len__(self):
        return len(self._settings.childKeys())

    def __contains__(self, key):
        return self._settings.contains(key)

    def keys(self):
        return self._settings.childKeys()

    def update(self, data):
        for key, value in list(data.items()):
            self[key] = value

    def clear(self):
        self._settings.remove('')

    def __repr__(self):
        return 'PySettings(%s)' % self._settings
